from to_sort.Reforged.Writer import Writer
import csv

class CsvWriter(Writer):
    """
    Implémentation du Writer pour écrire des données dans un fichier CSV.
    """
    def __init__(self, connection):
        """
        Initialise le Writer CSV.
        :param connection: Instance de la classe Connection (doit être de type 'file').
        """
        if connection.resource_type != 'file':
            raise ValueError("Le Writer CSV nécéssite une connexion de type 'file'.")
        super().__init__(connection)

    def write(self, data, mode: str = "insert", batch_size: int = None, if_exists: str = "none", if_not_exists: str = "fail"):
        """
        Écrit des données dans un fichier CSV.
        :param data: Données à écrire (liste de listes).
        :param mode: Mode d'écriture (seul 'insert' est supporté pour les CSV).
        :param batch_size: Taille des lots (non utilisé pour les CSV).
        :param if_exists: Action si le fichier existe (supprime, écrase, etc).
        :param if_not_exists: Action si le fichier n'existe pas.
        """
        file_path = self.connection.params.get("file_path")
        mode = "w" if if_exists == "deleteAndCreate" else "a"
        try:
            with open(file_path, mode=mode, newline='', encoding=self.connection.params.get("encoding", "utf-8")) as csv_file:
                writer = csv.writer(csv_file)
                if if_exists == "deleteAndCreate":
                    writer.writerow(data[0]) # Écrire l'en-tête si on recrée le fichier
                writer.writerow(data[1:])
        except Exception as e:
            raise RuntimeError(f"Erreur lors de l'écriture dans le fichier CSV: {e}")




from .timer import timer
from .logger import Logger
from .dbMysql import dbMysql
from .dbMssql import dbMssql
import polars as pl
import hashlib

class pipeline:
    SQL_TO_POLARS = {
        "int": pl.Int64,
        "integer": pl.Int64,
        "bigint": pl.Int64,
        "smallint": pl.Int32,
        "varchar": pl.Utf8,
        "text": pl.Utf8,
        "char": pl.Utf8,
        "time": pl.Time,
        "timestamp": pl.Datetime,
        "datetime": pl.Datetime,
        "datetime2": pl.Datetime,
        "date": pl.Date,
        "decimal": pl.Decimal,
        "numeric": pl.Decimal,
        "float": pl.Float64,
        "double": pl.Float64,
        "boolean": pl.Boolean,
    }

    def __init__(self, dictionnary: dict, logger: Logger):
        """
        Initialise un pipeline avec les informations de connexions aux bases de données
        :param dictionnary: le dictionnaire qui contient les informations du pipeline
            - 'db_source': la base de données source
            - 'query_source': la requête à envoyer à la source
            - 'tableau': les données sous forme de tableau (source alternative)
            - 'db_destination': la base de données destination
            - 'executemany' ou 'bulk'
            - 'batch_size': la taille des lots pour le traitement en batch
        :param logger: le logger pour gérer la journalisation des évènements du pipeline
        """
        self.logger = logger
        self.__db_source = dictionnary.get('db_source')
        self.__query_source = dictionnary.get('query_source')
        self.__tableau = dictionnary.get('tableau')
        self.db_destination = dictionnary.get('db_destination')
        self.operation_type = dictionnary.get('operation_type', 'insert')
        self.insert_method = dictionnary.get("insert_method", "bulk")
        self.batch_size = dictionnary.get('batch_size', 10_000)
        self.db_destination.get('db').batch_size = self.batch_size
        if self.__db_source:
            self.__db_source.batch_size = self.batch_size

    def _generate_hash(self, row, cols):
        """
        Génère un hash MD5 pour une ligne de valeurs donnée
        :param row: La ligne de données
        :param cols: Les colonnes de la table
        :return: Le hash de la ligne
        """
        concat_cols = "|".join(str(row[col]) if row[col] is not None else "null" for col in cols)
        return hashlib.md5(concat_cols.encode("utf-8")).hexdigest()

    def sql_defs_to_polars_schema(self, cols, cols_def, logger=None):
        schema = {}
        for col, sql_type in zip(cols, cols_def):
            base_type = sql_type.split("(")[0].lower()
            if isinstance(self.__db_source, dbMysql) and sql_type.lower() == 'time':
                pl_type = pl.Utf8
            else:
                pl_type = pipeline.SQL_TO_POLARS.get(base_type)
                if not pl_type:
                    if logger:
                        logger.debug(f"Type SQL non reconnu '{sql_type}', fallback en Utf8 pour la colonne '{col}'.")
                    pl_type = pl.Utf8
            schema[col] = pl_type
        return schema

    def _update_schema(self, old_schema, new_schema):
        """
        Met à jour l'ancien schéma en remplaçant les colonnes encore set à pl.Null
        :param old_schema: L'ancien schéma
        :param new_schema: Le nouveau schéma du batch en cours
        :return: Le schéma mis à jour
        """
        updated_schema = old_schema.copy()
        for col, dtype in old_schema.items():
            if dtype == pl.Null and col in new_schema:
                updated_schema[col] = new_schema[col]
        return updated_schema

    def _data_generator(self):
        """
        Générateur de données qui itère sur les données sources, qu'elles proviennent d'un tableau en mémoire
        ou d'une base de données, en les renvoyant sous forme de DataFrame par lots (batches).
        :return: Yield un DataFrame Polars contenant un batch de données.
        :raises ValueError: Si deux sources de données sont spécifiées (tableau et base de données)
        ou si aucune source de données valide n'est définie.
        """
        self.logger.info("Chargement des données depuis la source...")
        if self.__tableau is not None and self.__db_source is not None:
            msg = "Deux sources de données différentes sont définies, veuillez n'en choisir qu'une."
            self.logger.error(msg)
            raise ValueError(msg)
        cols = self.db_destination.get("cols", [])
        cols_def = self.db_destination.get("cols_def", [])
        hash_col = self.db_destination.get("hash")
        apply_hash = hash_col in cols if hash_col else False
        schema = None

        def process_batch(batch):
            nonlocal schema
            if isinstance(batch, pl.DataFrame):
                return batch
            batch_dicts = [dict(zip(cols, row)) for row in batch]
            if apply_hash:
                for row in batch_dicts:
                    cols_to_hash = [col for col in cols if col != hash_col]
                    row[hash_col] = self._generate_hash(row, cols_to_hash)
            if schema is None:
                if cols_def and all(t is not None for t in cols_def):
                    schema = self.sql_defs_to_polars_schema(cols, cols_def)
                    df = pl.DataFrame(batch_dicts, schema=schema, orient='row', strict=False)
                else:
                    self.logger.debug("Inférence des types sur le premier batch")
                    df = pl.DataFrame(batch_dicts, orient='row', strict=False, infer_schema_length=len(batch_dicts))
                    schema = df.schema
            else:
                if any(dtype == pl.Null for dtype in schema.values()):
                    self.logger.debug("Réinférence partielle des types")
                    df_temp = pl.DataFrame(batch_dicts, orient='row', strict=False, infer_schema_length=len(batch_dicts))
                    schema = self._update_schema(schema, df_temp.schema)
                df = pl.DataFrame(batch_dicts, schema=schema, orient='row', strict=False)
            return df

        if self.__tableau is not None and len(self.__tableau) > 0:
            for start in range(0, len(self.__tableau), self.batch_size):
                batch = self.__tableau[start:start + self.batch_size]
                try:
                    yield process_batch(batch)
                except Exception as e:
                    self.logger.error(f"Échec de la création du dataframe: {e}")
                    yield None, batch
        elif self.__db_source and self.__query_source:
            log_level, log_base = self.logger.disable()
            self.__db_source.connect()
            self.logger.enable(log_level, log_base)
            for batch in self.__db_source.sqlQuery(self.__query_source):
                try:
                    yield process_batch(batch)
                except Exception as e:
                    msg = f"Échec de la création du dataframe: {e}"
                    if isinstance(self.__db_source, dbMysql):
                        msg += "\n(Pour MySQL, pensez à utiliser TIME_FORMAT(col, '%H:%i:%s') sur vos colonnes TIME)"
                    elif isinstance(self.__db_source, dbMssql):
                        msg += "\n(Pour SQL Server, pensez à CAST(uuid_col AS NVARCHAR(36)) si vous avez des UUID en UNIQUEIDENTIFIER)"
                    for line in msg.split("\n("):
                        self.logger.error(line)
                    yield None, batch
        else:
            raise ValueError("Source de données non supportée.")

    @timer
    def create_destination_table(self, drop: bool=False):
        """Supprime et crée la table de destination du pipeline"""
        dst = self.db_destination.get('db')
        schema = self.db_destination.get('schema')
        table = self.db_destination.get('table')
        full_table = f"{schema}.{table}" if schema else table
        cols = self.db_destination.get('cols')
        cols_def = self.db_destination.get('cols_def')
        if len(cols) != len(cols_def):
            raise ValueError("Le nombre de colonnes (cols) ne correspond pas au nombre de définitions (cols_def)")
        columns = ', '.join([f"{col} {defn}" for col, defn in zip(cols, cols_def)])
        dst.connect()
        self.logger.info("Création de la table de destination...")
        try:
            if drop:
                dst.sqlExec(f"DROP TABLE IF EXISTS {full_table}")
            dst.sqlExec(f"CREATE TABLE {full_table} ({columns})")
            self.logger.info("Table de destination créée.")
        except Exception as e:
            self.logger.error(f"Erreur lors de la création de la table de destination: {e}")
            raise

    def _insert(self, rows):
        """Insertion par bulk ou executemany"""
        if self.insert_method == "bulk":
            return self.db_destination.get("db").insertBulk(
                schema=self.db_destination.get("schema"),
                table=self.db_destination.get("table"),
                rows=rows,
                cols=self.db_destination.get("cols")
            )
        elif self.insert_method == "executemany":
            return self.db_destination.get("db").insertMany(
                schema=self.db_destination.get("schema"),
                table=self.db_destination.get("table"),
                rows=rows,
                cols=self.db_destination.get("cols")
            )

    def _upsert(self, rows):
        """Upsert par bulk ou executemany"""
        if self.insert_method == "bulk":
            return self.db_destination.get("db").upsertBulk(
                schema=self.db_destination.get("schema"),
                table=self.db_destination.get("table"),
                rows=rows,
                cols=self.db_destination.get("cols"),
                conflict_cols=self.db_destination.get("conflict_cols")
            )
        elif self.insert_method == "executemany":
            return self.db_destination.get("db").upsertMany(
                schema=self.db_destination.get("schema"),
                table=self.db_destination.get("table"),
                rows=rows,
                cols=self.db_destination.get("cols"),
                conflict_cols=self.db_destination.get("conflict_cols")
            )

    def _perform_insertion(self, rows):
        """Effectue l'insertion ou l'upsert des données"""
        method_map = {
            "insert": self._insert,
            "upsert": self._upsert
        }
        return method_map[self.operation_type](rows)

    @timer
    def run(self):
        """
        Exécute le pipeline en insérant des données depuis la source vers la destination définie.
        :return: Une liste des lots rejetés contenant les erreurs lors de l'insertion.
        :raises Exception: Si une erreur autre qu'une erreur d'insertion survient pendant l'exécution du pipeline
        """
        rejects = []
        res = {"nb_lines_success": 0, "nb_lines_error": 0, "errors": rejects}
        batch_cpt = 1
        total = 0
        total_inserted = 0
        schema = self.db_destination.get('schema')
        table = self.db_destination.get('table')
        table_full = f"{schema}.{table}" if schema else table
        try:
            log_level, base_level = self.logger.disable()
            self.db_destination['db'].connect()
            self.logger.enable(log_level, base_level)
            name = self.db_destination.get('name', 'bdd')
            self.logger.info(f"Connexion à {name} réussie.")
            for batch_df in self._data_generator():
                taille = len(batch_df)
                total += taille
                if isinstance(batch_df, tuple) and batch_df[0] is None:
                    rejects.append((name, "Échec création dataframe", batch_df[1]))
                    res["nb_lines_error"] += len(batch_df[1])
                    continue
                rows = batch_df.rows()
                old_global, old_base = self.logger.disable()
                self.logger.enable(Logger.ERROR, Logger.ERROR)
                insert_result = self._perform_insertion(rows)
                self.logger.enable(old_global, old_base)
                if not insert_result:
                    raise ValueError("Aucune opération n'a été réalisée, vérifiez les operation_type et insert_method.")
                if insert_result[0] == "ERROR":
                    rejects.append((name, insert_result, rows))
                    res["nb_lines_error"] += taille
                else:
                    total_inserted += taille
                    res["nb_lines_success"] += taille
                    self.logger.info(
                        f"Batch {batch_cpt}: {taille} ligne(s) insérée(s) avec succès dans la table {table_full}. "
                        f"Total inséré: {total_inserted}/{total} ligne(s).")
                    batch_cpt += 1
        except Exception as e:
                self.logger.enable()
                self.logger.error(f"Échec de l'exécution du pipeline: {e}")
                raise
        return res


from abc import ABC, abstractmethod

class Reader(ABC):
    """
    Interface de base pour la lecture des données depuis une connexion.
    """
    def __init__(self, connection):
        """
        Initialise le Reader avec une connexion donnée.
        :param connection: Instance de la classe Connection utilisée pour accéder aux données.
        """
        self.connection = connection

    @abstractmethod
    def read(self, query_or_path: str, batch_size: int = None):
        """
        Lit les données depuis la ressource.
        :param query_or_path: Requête SQL ou chemin de fichier.
        :param batch_size: Taille des données à lire par lot (None pour tout lire).
        :return: Données lues (liste, DataFrame, etc.).
        """
        pass


import psycopg2
from to_sort.Reforged.Connection import Connection

class PostgreSqlConnection(Connection):
    """
    Implémentation de l'interface Connection pour PostgreSQL.
    """
    def __init__(self, **kwargs):
        """
        Initialise la connexion PostgreSQL.
        :param kwargs: Paramètres spécifiques à la connexion (hôte, utilisateur, etc.).
        """
        super().__init__(resource_type="sql", **kwargs)

    def open(self):
        if self.connection:
            raise RuntimeError("La connexion est déjà ouverte.")
        try:
            self.connection = psycopg2.connect(
                host=self.params.get("host"),
                user=self.params.get("user"),
                password=self.params.get("password"),
                dbname=self.params.get("database"),
                port=self.params.get("port"),
                connect_timeout=self.params.get("timeout", 10)
            )
        except Exception as e:
            raise RuntimeError(f"Erreur lors de l'ouverture de la connexion PostgreSQL : {e}")

    def close(self):
        if self.connection:
            self.connection.close()
            self.connection = None

    def commit(self):
        if self.connection:
            try:
                self.connection.commit()
            except Exception as e:
                raise RuntimeError(f"Erreur lors du commit PostgreSQL : {e}")
        else:
            raise RuntimeError("Impossible de commit : aucune connexion active.")

    def is_open(self) -> bool:
        return self.connection is not None and self.connection.closed == 0


from to_sort.Reforged.Reader import Reader
import csv

class CsvReader(Reader):
    """
    Implémentation du Reader pour lire des données depuis un fichier CSV.
    """
    def __init__(self, connection):
        """
        Initialise le Reader CSV.
        :param connection: Instance de la classe Connection (doit être de type 'file').
        """
        if connection.resource_type != "file":
            raise ValueError("Le Reader CSV nécessite une connexion de type 'file'.")
        super().__init__(connection)

    def read(self, file_path: str, batch_size: int = None):
        """
        Lit les données d'un fichier CSV.
        :param file_path: Chemin du fichier CSV.
        :param batch_size: Nombre de lignes à lire par lot (None pour tout lire).
        :return: Un générateur qui retourne les données par lots ou en entier.
        """
        self.connection.logger.debug(f"Lecture du fichier CSV : {file_path}")
        try:
            with open(file_path, mode='r', encoding=self.connection.params.get("encoding", "utf-8")) as csv_file:
                reader = csv.reader(csv_file)
                header = next(reader)  # Lire l'en-tête
                if batch_size:
                    batch = []
                    for row in reader:
                        batch.append(row)
                        if len(batch) == batch_size:
                            yield [header] + batch
                            batch = []
                    if batch:
                        yield [header] + batch
                else:
                    yield [header] + list(reader)
        except Exception as e:
            self.connection.logger.error(f"Erreur lors de la lecture du fichier CSV : {e}")
            raise


from abc import ABC, abstractmethod

class Writer(ABC):
    """
    Interface de base pour écrire des données dans une ressource.
    """
    def __init__(self, connection):
        """
        Initialise le Writer avec une connexion donnée.
        :param connection: Instance de la classe Connection utilisée pour accéder à la ressource.
        """
        self.connection = connection

    @abstractmethod
    def write(self, data, mode: str = "insert", batch_size: int = None, if_exists: str = "none",
              if_not_exists: str = "fail"):
        """
        Écrit des données dans la ressource.
        :param data: Données à écrire (liste de tuples, DataFrame, etc.).
        :param mode: Mode d'écriture (insert, update, delete, etc.).
        :param batch_size: Taille des données à écrire par lot (None pour tout écrire d'un coup).
        :param if_exists: Action si la ressource existe (deleteAndCreate, deleteRows, truncate, fail, none, syncSchema).
        :param if_not_exists: Action si la ressource n'existe pas (create, fail).
        """
        pass


import os
import time
import timeit
from .global_config import get_timer

def timer(func):
    """
    Décorateur pour mesurer le temps d'exécution d'une fonction et enregistrer ce temps dans un logger si activé.
    :param func: La fonction à décorer pour mesurer et enregistrer son temps d'exécution.
    :raises ValueError: Si aucun logger n'est défini dans les arguments de la fonction appelée.
    :return: La fonction décorée qui mesure le temps d'exécution.
    """
    def wrapper(*args, **kwargs):
        if get_timer():
            logger = kwargs.get('logger', None)
            if logger is None and args and hasattr(args[0], 'logger'):
                logger = args[0].logger
            if logger is None: raise ValueError("Pas de logger défini.")
            start_time = timeit.default_timer()
            result = func(*args, **kwargs)
            elapsed_time = timeit.default_timer() - start_time
            if logger is not None:
                logger.info(f"Temps d'exécution de {func.__name__}: {elapsed_time:.4f} secondes.")
            return result
        else:
            return func(*args, **kwargs)
    return wrapper

def set_timezone(tz='Europe/Paris'):
    """Définit le fuseau horaire global du processus"""
    os.environ['TZ'] = tz
    if hasattr(time, 'tzset'):
        time.tzset()
        print(f"[INFO] Fuseau horaire défini globalement sur {tz} (via time.tzset)")
    else:
        from zoneinfo import ZoneInfo
        global _GLOBAL_TZ
        _GLOBAL_TZ = ZoneInfo(tz)
        print(f"[WARNING] Fuseau horaire simulé sur {tz} (Windows n’a pas tzset())")
        print(f"Appeler ads.now() pour avoir l'heure simulée sur {tz}")

def now():
    """Renvoie l'heure actuelle dans le fuseau global défini"""
    from datetime import datetime
    if '_GLOBAL_TZ' in globals():
        return datetime.now(_GLOBAL_TZ)
    return datetime.now()


import os
from .logger import Logger
from git import Repo
from github import Github
from .timer import timer

class GitHandler:
    def __init__(self, token: str, logger: Logger):
        self.token = token
        self.github = Github(token)
        self.logger = logger

    def check_permissions(self, path: str):
        if not os.access(path, os.R_OK):
            raise PermissionError(f"Le script n'a pas la permission de lecture sur le chemin: {path}")

    @timer
    def clone_or_update(self, destination_path: str, repository: str, branch: str = None):
        """Clone le repo s'il existe pas localement, sinon pull la dernière version."""
        self.logger.info(f"Connexion au repo {repository}")
        url = f"https://{self.token}@github.com/{repository}.git"
        try:
            if os.path.exists(destination_path):
                self.check_permissions(destination_path)
                repo = Repo(destination_path)

                self.logger.debug("Mise à jour du remote origin avec le token.")
                repo.remotes.origin.set_url(url)

                repo.remotes.origin.fetch()
                if branch:
                    self.logger.debug(f"Changement de branche : {branch}")
                    repo.git.checkout(branch)
                repo.remotes.origin.pull()
                self.logger.info(f"Le repo {repository} a été mis à jour.")
            else:
                self.logger.info(f"Clonage du repo {repository} dans {destination_path}...")
                if branch:
                    Repo.clone_from(url, destination_path, branch=branch)
                else:
                    Repo.clone_from(url, destination_path)
                self.logger.info(f"Le repo {repository} a été cloné avec succès.")
        except Exception as e:
            self.logger.error(f"Erreur lors du clonage du repo {repository}: {e}")
            raise

    @timer
    def setup_virtualenv(self, repo_path: str):
        """Création et installation d'un environnement virtuel dans le repo."""
        venv_path = os.path.join(repo_path, "venv")
        requirements_path = os.path.join(repo_path, "requirements.txt")
        launch_script = os.path.join(repo_path, "launch.sh")
        self.logger.info(f"Création de l'environnement virtuel dans {venv_path}...")
        os.system(f"python -m venv {venv_path}")
        if os.path.exists(requirements_path):
            self.logger.info("Installation des dépendances à partir de requirements.txt...")
            activate = os.path.join(venv_path, "Scripts",
                                    "activate") if os.name == "nt" else f"source {venv_path}/bin/activate"
            install_cmd = f"{activate} && pip install -r {requirements_path}"
            os.system(install_cmd)
        else:
            self.logger.warning("Aucun fichier requirements.txt trouvé, installation ignorée.")
        # Attribution des droits d’exécution au script launch.sh (si présent)
        if os.path.exists(launch_script):
            self.logger.info("Attribution des droits d'exécution sur launch.sh...")
            try:
                os.chmod(launch_script, 0o755)
                self.logger.info("Droits d'exécution appliqués avec succès sur launch.sh.")
            except Exception as e:
                self.logger.error(f"Impossible de modifier les permissions de launch.sh : {e}")
        else:
            self.logger.warning("Aucun fichier launch.sh trouvé dans le dépôt.")
        self.logger.info("Environnement virtuel configuré avec succès.")


from to_sort.Reforged.ReaderHelper import ReaderHelper

class SqlReaderHelper(ReaderHelper):
    """
    Implémentation de ReaderHelper pour une base SQL.
    """
    def get_columns(self, source: str):
        """
        Récupère les colonnes d'une table ou d'une requête SQL.
        :param source: Nom de la table ou requête SQL.
        :return: Liste de tuples (nom_colonne, type_colonne).
        """
        query = f"""
        SELECT COLUMN_NAME, DATA_TYPE
        FROM INFORMATION_SCHEMA.TABLES
        WHERE TABLE_NAME = '{source}'
        """
        try:
            cursor = self.connection.connection.cursor()
            cursor.execute(query)
            return cursor.fetchall()
        except Exception as e:
            raise RuntimeError(f"Erreur lors de la récupération des colonnes: {e}")

    def guess_read(self, source: str):
        """
        Devine les paramètres nécéssaires pour lire les données d'une source SQL.
        :param source: Nom de la table SQL ou requête SQL.
        :return: Structure devinée des données (colonnes, types...)
        """
        columns = self.get_columns(source)
        return {
            "columns": [col[0] for col in columns],
            "types": [col[1] for col in columns]
        }


class dbSql:
    def __init__(self, dictionnary: dict, logger, batch_size=10_000):
        """
        instancie la classe dbSql, qui permet la communication avec une base de données
        :param dictionnary: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion MySQL
        :param logger: Un logger ads qui va gérer les logs des actions de la classe
        :param batch_size: La taille des batchs en lecture et écriture
        """
        self.connection = None
        self.logger = logger
        self.__techno = dictionnary.get('techno')
        self.__database = dictionnary.get('database')
        self.__user = dictionnary.get('user')
        self.__password = dictionnary.get('password')
        self.__port = dictionnary.get('port')
        self.__host = dictionnary.get('host')
        self.__charset = dictionnary.get('charset', 'UTF-8')
        self.batch_size = batch_size
        self.__package = dictionnary.get('package', 'pymssql')


from abc import ABC, abstractmethod

class WriterHelper(ABC):
    """
    Interface pour aider à paramètrer un Writer
    """
    def __init__(self, connection):
        """
        Initialise le WriterHelper avec une connexion donnée.
        :param connection: Instance de la classe connection utilisée pour accèder aux données.
        """
        self.connection = connection

    @abstractmethod
    def guess_insert(self, target: str, columns: list):
        """
        Génère automatiquement une requête d'insertion.
        :param target: Nom de la table ou du fichier cible.
        :param columns: Liste des colonnes.
        :return: Requête ou commande d'insertion
        """
        pass

    @abstractmethod
    def guess_update(self, target: str, columns: list, conflict_cols: list):
        """
        Génère automatiquement une requête de mise à jour.
        :param target: Nom de la table ou du fichier cible.
        :param columns: Liste des colonnes à mettre à jour.
        :param conflict_cols: Colonnes utilisées pour détecter les conflits.
        :return: Requête de mise à jour.
        """
        pass

    @abstractmethod
    def guess_if_exists_delete_and_create(self, target: str):
        """
        Génère automatiquement les commandes pour supprimer et recréer une ressource.
        :param target: Nom de la table ou du fichier cible.
        :return: Commandes pour supprimer et recréer.
        """
        pass


import polars as pl
from .timer import timer
from .logger import Logger
from polars.testing import assert_frame_equal

class DataComparator:
    def __init__(self, dictionnary: dict, logger: Logger):
        """
        Initialise un dataComparator avec les informations de connexions aux bases de données
        :param dictionnary: le dictionnaire qui contient les informations du pipeline
            - 'db_source_1': la base de données source 1
            - 'query_source_1': la requête à envoyer à la source 1
            - 'db_source_2': la base de données source 2
            - 'query_source_2': la requête à envoyer à la source 2
            - 'batch_size': la taille des lots pour le traitement en batch
        :param logger: le logger pour gérer la journalisation des évènements du pipeline
        """
        self.logger = logger
        self.__db_source_1 = dictionnary.get('db_source_1')
        self.__query_source_1 = dictionnary.get('query_source_1')
        self.__db_source_2 = dictionnary.get('db_source_2')
        self.__query_source_2 = dictionnary.get('query_source_2')
        self.__batch_size = dictionnary.get('batch_size', 10_000)

    def _fetch_data(self, db, query):
        """
        Exécute la requête donnée sur la base spécifiée et retourne un DataFrame Polars
        :param db: la base sur laquelle envoyer la requête
        :param query: la requête à envoyer
        :return: le résultat de la requête sous forme de DataFrame
        """
        db.connect()
        for batch in db.sqlQuery(query):
            yield pl.DataFrame(batch, orient='row', strict=False, infer_schema_length=self.__batch_size)

    @timer
    def compare(self, check_dtypes: bool = False):
        """
        Compare les résultats des deux requêtes SQL entre les deux bases de données.

        :param check_dtypes: si True, vérifie aussi la stricte égalité des types de colonnes
                             si False (défaut), ne vérifie que les valeurs
        :return: True si les DataFrames sont identiques (selon le niveau de vérification)
        """
        try:
            gen1 = self._fetch_data(self.__db_source_1, self.__query_source_1)
            gen2 = self._fetch_data(self.__db_source_2, self.__query_source_2)
            batch_index = 0
            while True:
                try:
                    df1 = next(gen1)
                except StopIteration:
                    df1 = None
                try:
                    df2 = next(gen2)
                except StopIteration:
                    df2 = None
                if df1 is None and df2 is None:
                    self.logger.info("Les données des deux sources sont identiques.")
                    return True
                if (df1 is None) != (df2 is None):
                    self.logger.error(f"Nombre de lignes différent entre les deux sources.")
                    return False
                batch_index += 1
                try:
                    assert_frame_equal(df1, df2, check_dtypes=check_dtypes)
                except Exception as e:
                    self.logger.error(f"Les données diffèrent dans le batch #{batch_index}: {e}")
                    return False
                self.logger.debug(f"Batch #{batch_index} identique ({len(df1)} lignes).")
        except Exception as e:
            self.logger.error(f"Erreur lors de la comparaison des données : {e}")
            raise


import os
import dotenv
import inspect

class env:
    def __init__(self, logger, file=None):
        """
        Recherche et charge un fichier .env.
        - Si 'file' est fourni : charge ce fichier s'il existe.
        - Sinon : cherche automatiquement le .env le plus proche en remontant
          depuis le dossier du script exécuté.
        """
        self.logger = logger

        # --- Cas 1 : fichier spécifié explicitement
        if file and os.path.isfile(file):
            dotenv_file = os.path.abspath(file)
            dotenv.load_dotenv(dotenv_file, override=True)
            self.logger.debug(f"Fichier .env spécifié trouvé et chargé : {dotenv_file}")

        # --- Cas 2 : recherche automatique
        else:
            caller_file = inspect.stack()[1].filename
            caller_dir = os.path.dirname(os.path.abspath(caller_file))
            dotenv_file = self._find_env_upwards(caller_dir)
            if dotenv_file:
                dotenv.load_dotenv(dotenv_file, override=True)
                self.logger.debug(f"Fichier .env trouvé automatiquement : {dotenv_file}")
            else:
                self.logger.warning("Aucun fichier .env trouvé en remontant depuis le script.")
                self.logger.warning("Veuillez créer un fichier .env à la racine du projet ou en spécifier un manuellement.")

        # --- Synchronisation des variables dans l'objet
        for cle, valeur in os.environ.items():
            setattr(self, cle, valeur)

    def _find_env_upwards(self, start_path):
        """Recherche récursivement un fichier .env en remontant les dossiers"""
        current_dir = start_path
        while True:
            self.logger.debug(f"Recherche d'un fichier .env dans : {current_dir}")
            candidate = os.path.join(current_dir, '.env')
            if os.path.isfile(candidate):
                return os.path.abspath(candidate)
            parent_dir = os.path.dirname(current_dir)
            if parent_dir == current_dir:
                break
            current_dir = parent_dir
        return None


from .mail import mail
from .timer import now
from .loadEnv import env
from .logger import Logger
from .dbPgsql import dbPgsql
from .dbMssql import dbMssql
from .dbMysql import dbMysql
from .pipeline import pipeline
from .odoo import OdooConnector
from .cdc import ChangeDataCapture
from .git_handler import GitHandler
from .file_handler import FileHandler
from .dataComparator import DataComparator
from .timer import timer, get_timer, set_timezone
from .global_config import set_timer, get_public_ip

__all__ = [
    "Logger", "timer", "set_timer", "get_timer", "dbPgsql", "dbMssql", "pipeline", "OdooConnector",
    "env", "DataComparator", "dbMysql", "mail", "FileHandler", "ChangeDataCapture",
    "set_timezone", "GitHandler", "now", "get_public_ip"
]


import xmlrpc.client

class OdooConnector:
    def __init__(self, dictionnary: dict, logger):
        """
        Classe pour se connecter à Odoo via xmlrpc et exécuter des opérations CRUD
        :param dictionnary: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion Odoo
        :param logger: Un logger ads qui va gérer les logs des actions de la classe
        """
        self.models = None
        self.uid = None
        self.common = None
        self.name = dictionnary['name']
        self.url = dictionnary['url']
        self.db = dictionnary['db']
        self.user = dictionnary['user']
        self.password = dictionnary['password']
        self.logger = logger

    def connect(self):
        """
        Lance la connection à Odoo
        """
        try:
            self.logger.info("Tentative de connexion à Odoo...")
            self.common = xmlrpc.client.ServerProxy('{}/xmlrpc/common'.format(self.url))
            self.uid = self.common.authenticate(self.db, self.user, self.password, {})
            if not self.uid:
                msg = "Échec de l'authentification, identifiants incorrects."
                self.logger.error(msg)
                raise Exception(msg)
            self.models = xmlrpc.client.ServerProxy('{}/xmlrpc/2/object'.format(self.url))
            self.logger.info(f"Connexion à Odoo réussie.")
            self.logger.debug(f"Connecté en tant que {self.user} (UID: {self.uid})")
        except Exception as e:
            self.logger.error(f"Erreur lors de la connexion à Odoo : {e}")
            raise

    def __str__(self):
        """
        Retourne le nom de la connexion Odoo
        """
        return f"Connexion Odoo: {self.name}"

    def desc(self, table: str):
        """
        Retourne la description des champs d'une table Odoo
        :param table: Le nom du modèle
        :return: Un dictionnaire qui contient les informations sur les champs
        """
        try:
            self.logger.info(f"Récupération des informations de la table: {table}")
            fields = self.models.execute_kw(
                self.db, self.uid, self.password, table, 'fields_get', [], {'attributes': ['string', 'help', 'type']}
            )
            self.logger.info(f"Descriptions récupérées pour la table: {table}")
            return fields
        except Exception as e:
            self.logger.error(f"Erreur lors de la récupération des descriptions: {e}")
            raise

    def get(self, table, fields, filter):
        """
        Récupère des enregistrements dans Odoo avec un filtre
        :param table: Le nom du modèle Odoo
        :param fields: Une liste de champs à récupérer
        :param filter: Un filtre Odoo
        :return: Une liste d'enregistrements correspondant au filtre
        """
        try:
            self.logger.info(f"Récupération des enregistrements depuis la table {table} avec le filtre: {filter}")
            records = self.models.execute_kw(
                self.db, self.uid, self.password, table, 'search_read', [filter], {'fields': fields}
            )
            self.logger.info(f"{len(records)} enregistrement(s) récupérés depuis {table}")
            return records
        except Exception as e:
            self.logger.error(f"Erreur lors de la récupération des enregistrements depuis {table}: {e}")
            raise

    def put(self, table: str, lines: []):
        """
        Insère de nouveaux enregistrements dans Odoo
        :param table: Le nom du modèle où insérer
        :param lines: Une liste contenant les données à insérer
        :return: Une liste d'IDs des enregistrements crées
        """
        try:
            self.logger.info(f"Insertion de {len(lines)} enregistrements dans {table}")
            ids = []
            for line in lines:
                id = self.models.execute_kw(
                    self.db, self.uid, self.password, table, 'create', [line]
                )
                ids.append(id)
                self.logger.info(f"Enregistrement(s) inséré(s) avec succès, ID: {id}")
            return ids
        except Exception as e:
            self.logger.error(f"Erreur lors de l'insertion dans {table}: {e}")
            raise



from to_sort.Reforged.Writer import Writer

class SqlWriter(Writer):
    """
    Implémentation du Writer pour écrire des données dans une base SQL.
    """
    def __init__(self, connection):
        """
        Initialise le Writer SQL.
        :param connection: Instance de la classe Connection SQL
        """
        super().__init__(connection)

    def write(self, data, mode: str = "insert", batch_size: int = None, if_exists: str = "none", if_not_exists: str = "fail"):
        """
        Écrit des données dans une base SQL.
        :param data: Données à écrire (liste de tuples).
        :param mode: Mode d'écriture (insert, update, delete, etc.).
        :param batch_size: Taille des lots.
        :param if_exists: Action si la ressource existe.
        :param if_not_exists: Action si la ressource n'existe pas.
        :return:
        """
        if not self.connection.is_open():
            raise RuntimeError("La connexion doit être ouverte avant d'écrire des données.")
        self._handle_resource(if_exists, if_not_exists)
        if mode == "insert":
            self._write_insert(data, batch_size)
        elif mode == "update":
            self._write_update(data, batch_size)
        elif mode == "delete":
            self._write_delete(data, batch_size)
        else:
            raise ValueError(f"Mode d'écriture non supporté: {mode}")

    def _handle_resource(self, if_exists, if_not_exists):
        """
        Gère les actions à effectuer sur la resource (table) en fonction des options 'if_exists' et 'if_not_exists'
        """
        cursor = self.connection.connection.cursor()
        if if_not_exists == "create":
            cursor.execute("CREATE TABLE IF NOT EXISTS ...")
            self.connection.commit()
        elif if_not_exists == "fail":
            cursor.execute("SELECT 1 FROM INFORMATION_SCHEMA.TABLES WHERE ...")
            if not cursor.fetchone():
                raise RuntimeError("La table n'existe pas.")
        if if_exists == "deleteAndCreate":
            cursor.execute("DROP TABLE IF EXISTS ...")
            cursor.execute("CREATE TABLE ...")
            self.connection.commit()
        elif if_exists == "truncate":
            cursor.execute("TRUNCATE TABLE ...")
            self.connection.commit()

    def _write_insert(self, data, batch_size):
        """
        Gère l'insertion des données.
        """
        cursor = self.connection.connection.cursor()
        for batch in self._batch_data(data, batch_size):
            query = "INSERT INTO ... VALUES (%s, %s, ...)"
            cursor.executemany(query, batch)
            self.connection.commit()

    def _write_update(self, data, batch_size):
        """
        Gère la mise à jour des données.
        """
        cursor = self.connection.connection.cursor()
        for batch in self._batch_data(data, batch_size):
            query = "UPDATE ... SET col1 = %s, col2 = %s WHERE ..."
            cursor.executemany(query, batch)
            self.connection.commit()

    def _write_delete(self, data, batch_size):
        """
        Gère la suppression des données.
        """
        cursor = self.connection.connection.cursor()
        for batch in self._batch_data(data, batch_size):
            query = "DELETE FROM ... WHERE ..."
            cursor.executemany(query, batch)
            self.connection.commit()

    @staticmethod
    def _batch_data(data, batch_size):
        """
        Divise les données en lots pour éviter les surcharges mémoires
        """
        if batch_size:
            for i in range(0, len(data), batch_size):
                yield data[i:i + batch_size]
        else:
            yield data


import pymssql
import psycopg2
import os
import pymysql
import timeit
import polars as pl

from io import StringIO

from .timer import timer, get_timer
from .dataFactory import data_factory

class dbSql(data_factory):
    def __init__(self, dictionnary: dict, logger, batch_size=10_000):
       """
       instancie la classe dbSql, qui hérite de dataFactory
       :param dictionnary: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion sql server
       :param logger: un logger ads qui va gérer les logs des actions de la classe
       :param batch_size: la taille des batchs en lecture et écriture
       :param package: le package à utiliser ('pymssql' qui gère les connexion encryptées mais qui encode mal des
           caractères spéciaux en bulk ou alors 'pytds' qui est moins permissif et ne gère pas les connexion encryptées
           mais n'encode pas mal les caractères spéciaux
       """
       self.connection = None
       self.logger = logger
       self.__techno = dictionnary.get('techno')
       self.__database = dictionnary.get('database')
       self.__user = dictionnary.get('user')
       self.__password = dictionnary.get('password')
       self.__port = dictionnary.get('port')
       self.__host = dictionnary.get('host')
       self.__charset = dictionnary.get('charset', 'UTF-8')
       self.batch_size = batch_size

    def __connect_mssql(self):
        if self.logger is not None: self.logger.info("Tentative de connexion avec la base de données.")
        try:
            server = f"{self.__host}:{self.__port}" if self.__port else self.__host
            self.connection = pymssql.connect(
                server=server,
                user=self.__user,
                password=self.__password,
                database=self.__database,
                charset=self.__charset
            )
            if self.logger is not None: self.logger.info("Connexion établie avec la base de données.")
            return self.connection
        except Exception as e:
            if self.logger is not None: self.logger.error(f"Échec de la connexion à la base de données: {e}")
            raise

    def __connect_pg(self):
        if self.logger is not None: self.logger.info("Tentative de connexion avec la base de données.")
        try:
            self.connection = psycopg2.connect(
                database=self.__database,
                user=self.__user,
                password=self.__password,
                port=self.__port,
                host=self.__host
            )
            if self.logger is not None: self.logger.info(f"Connexion établie avec la base de données.")
            return self.connection
        except Exception as e:
            if self.logger is not None: self.logger.error(f"Échec de la connexion à la base de données: {e}")
            raise

    def __connect_mysql(self, additionnal_parameters=None):
        if self.logger is not None: self.logger.info("Tentative de connexion avec la base de données.")
        try:
            self.connection = pymysql.connect(
                host=self.__host,
                port=int(self.__port),
                user=self.__user,
                password=self.__password,
                database=self.__database,
                autocommit=False,
                local_infile=True
            )
            if additionnal_parameters:
                cursor = self.connection.cursor()
                for param, value in additionnal_parameters.items():
                    cursor.execute(f"SET SESSION {param} = {value}")
                self.connection.commit()
            if self.logger is not None: self.logger.info("Connexion établie avec la base de données.")
            return self.connection
        except Exception as e:
            if self.logger is not None: self.logger.error(f"Échec de la connexion à la base de données: {e}")
            raise

    @timer
    def connect(self):
        """
        lance la connexion avec les identifiants passés à l'initialisation de la classe
        toutes les méthodes de la classe nécéssitent une connexion active
        :return: la connexion
        """
        if self.__techno == "mssql":
            connexion = self.__connect_mssql()
        elif self.__techno == "postgresql":
            connexion = self.__connect_pg()
        elif self.__techno == "mysql":
            connexion = self.__connect_mysql()
        else:
            msg = f"La connexion {self.__techno} n'est pas supportée. Tentez avec 'mssql', 'postgresql' ou 'mysql'"
            self.logger.error(msg)
            raise ValueError(msg)
        return connexion

    @timer
    def sqlQuery(self, query):
        """
        lit la base de données avec la requête query
        :param query: la requête
        :return: les données lues avec yield
        """
        self.logger.debug(f"Exécution de la requête de lecture : {query}")
        try:
            timer_start = timeit.default_timer()
            cpt_rows = 0
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                self.logger.info("Requête exécutée avec succès, début de la lecture des résultats.")
                while True:
                    rows = cursor.fetchmany(self.batch_size)
                    if not rows:
                        break
                    yield rows
                    cpt_rows += len(rows)
                    self.logger.info(f"{cpt_rows} ligne(s) lue(s).")
            if get_timer():
                elapsed_time = timeit.default_timer() - timer_start
                self.logger.info(f"Temps d'exécution de sqlQuery: {elapsed_time:.4f} secondes")
            return cpt_rows
        except Exception as e:
            self.logger.error(f"Échec de la lecture des données: {e}")
            raise

    @timer
    def sqlExec(self, query):
        """
        execute une requête sur la base de données, un create ou delete table par exemple
        :param query: la requête
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                self.connection.commit()
                self.logger.info(f"Requête exécutée avec succès.")
        except Exception as e:
            self.logger.error(f"Échec de l'exécution de la requête: {e}")
            raise

    @timer
    def sqlScalaire(self, query):
        """
        execute une requête et retourne le premier résultat
        :param query: la requête
        :return: le résultat de la requête
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                result = cursor.fetchone()
                self.logger.info(f"Requête scalaire exécutée avec succès.")
                return result[0] if result else None
        except Exception as e:
            self.logger.error(f"Échec de l'exécution de la requête: {e}")
            raise

    @timer
    def insert(self, schema: str, table: str, cols: [], row: []):
        """
        Insère des données dans la base de données
        :param schema: Nom du schéma dans lequel est la table
        :param table: Nom de la table dans laquelle insérer
        :param cols: liste des colonnes dans lesquelles insérer
        :param row: liste des valeurs à insérer
        :return: le résultat de l'opération, l'erreur et la la ligne concernée en cas d'erreur
        """
        placeholders = ", ".join(["%s"] * len(cols))
        table = f"{schema}.{table}" if schema else table
        query = f"INSERT INTO {table} ({', '.join(cols)}) VALUES ({placeholders})"
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeur(s) insérée(s) avec succès dans la table {table}")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Échec de l'insertion des données: {e}")
            return "ERROR", str(e), row

    @timer
    def insertMany(self, schema: str, table: str, cols: [], rows: [[]]):
        """
        Insère des données par batch dans une table avec gestion des erreurs

        :param schema: Nom du schéma dans lequel se trouve la table
        :param table: Nom de la table dans laquelle insérer
        :param cols: Liste des colonnes dans lesquelles insérer
        :param rows: Liste des lignes à insérer
        :return: Le résultat de l'opération, l'erreur et la le batch concerné en cas d'erreur
        """
        failed_batches = []
        errors = []
        total_inserted = 0
        placeholders = ', '.join(["%s"] * len(cols))
        table = f"{schema}.{table}" if schema else table
        query = f"INSERT INTO {table} ({', '.join(cols)}) VALUES ({placeholders})"
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) avec succès dans la table {table}. "
                            f"Total inséré: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
                return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows

    def __get_df(self, rows, cols):
        if isinstance(rows, list):
            df = pl.DataFrame(rows, schema=cols, orient='row', infer_schema_length=10_000)
        elif isinstance(rows, pl.DataFrame):
            df = rows
        else:
            raise ValueError("Les données doivent être une liste de tuples ou un DataFrame polars")
        return df

    def __insertBulk_mssql(self, schema: str, table: str, cols: [], rows):
        import json
        import datetime
        failed_batches = []
        errors = []
        total_inserted = 0
        table_full = f"{schema}.{table}" if schema else table
        df = self.__get_df(rows, cols)
        type_mapping = {}
        for col in cols:
            non_null_values = df[col].drop_nulls()
            if len(non_null_values) > 0:
                sample_value = non_null_values[0]
            else:
                sample_value = 0
            if isinstance(sample_value, datetime.date):
                type_mapping[col] = "DATETIME2"
            elif isinstance(sample_value, datetime.time):
                type_mapping[col] = "TIME(3)"
            else:
                type_mapping[col] = "NVARCHAR(MAX)"
        col_defs = ", ".join([f"[{col}]" for col in cols])
        openjson_defs = ", ".join([f"[{col}] {type_mapping[col]} '$.{col}'" for col in cols])
        query = f"""
                INSERT INTO {table_full} ({col_defs})
                SELECT {col_defs}
                FROM OPENJSON(%s)
                WITH ({openjson_defs});
                """
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, df.height, self.batch_size), start=1):
                    batch = df.slice(start, self.batch_size)
                    try:
                        as_json = [row for row in batch.iter_rows(named=True)]
                        json_data = json.dumps(as_json, default=str)
                        cursor.execute(query, (json_data,))
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) avec succès dans la table {table_full}. "
                            f"Total inséré: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
            return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows

    def __insertBulk_pg(self, schema: str, table: str, cols: [], rows):
        failed_batches = []
        errors = []
        total_inserted = 0
        try:
            df = self.__get_df(rows, cols)
            table = f"{schema}.{table}" if schema else table
            n_rows = df.shape[0]
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, n_rows, self.batch_size), start=1):
                    batch = df.slice(start, self.batch_size)
                    try:
                        buffer = StringIO()
                        batch.write_csv(buffer, include_header=False)
                        buffer.seek(0)
                        columns = ', '.join(cols)
                        query = f"COPY {table} ({columns}) FROM STDIN WITH CSV DELIMITER ','"
                        cursor.copy_expert(query, buffer)
                        self.connection.commit()
                        buffer.close()
                        total_inserted += batch.shape[0]
                        self.logger.info(f"Batch {batch_index}: {batch.shape[0]} ligne(s) insérée(s) avec succès dans la table {table}. "
                                         f"Total inséré: {total_inserted}/{n_rows} ligne(s).")
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
            return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows

    def __insertBulk_mysql(self, schema: str, table: str, cols: [], rows):
        failed_batches = []
        errors = []
        total_inserted = 0
        table_full = f"{schema}.{table}" if schema else table
        temp_file = "temp_bulk_insert.csv"
        query = f"""
        LOAD DATA LOCAL INFILE '{temp_file}'
        INTO TABLE {table_full}
        FIELDS TERMINATED BY ','
        LINES TERMINATED BY '\n'
        ({', '.join(cols)})
        """
        try:
            df = self.__get_df(rows, cols)
            n_rows = df.shape[0]
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, n_rows, self.batch_size), start=1):
                    batch = df.slice(start, self.batch_size)
                    try:
                        batch.write_csv(temp_file, include_header=False)
                        cursor.execute(query)
                        self.connection.commit()
                        os.remove(temp_file)
                        total_inserted += batch.shape[0]
                        self.logger.info(
                            f"Batch {batch_index}: {batch.shape[0]} ligne(s) insérée(s) avec succès dans la table {table_full}. "
                            f"Total inséré: {total_inserted}/{n_rows} ligne(s).")
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
            return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows
        finally:
            if os.path.exists(temp_file):
                os.remove(temp_file)

    @timer
    def insertBulk(self, schema: str, table: str, cols: [], rows):
        """
        Insère des données dans une table en bulk, en lots

        :param schema: Nom du schéma
        :param table: Nom de la table dans laquelle insérer
        :param cols: Liste des colonnes dans lesquelles insérer
        :param rows: Liste des lignes à insérer
        :return: Le résultat de l'opération, l'erreur et la le batch concerné en cas d'erreur
        """
        if self.__techno == "mssql":
            result, errors, failed_batches = self.__insertBulk_mssql(schema, table, cols, rows)
        elif self.__techno == "postgresql":
            result, errors, failed_batches = self.__insertBulk_pg(schema, table, cols, rows)
        elif self.__techno == "mysql":
            result, errors, failed_batches = self.__insertBulk_mysql(schema, table, cols, rows)
        else:
            msg = f"La connexion {self.__techno} n'est pas supportée. Tentez avec 'mssql', 'postgresql' ou 'mysql'"
            self.logger.error(msg)
            raise ValueError(msg)
        return result, errors, failed_batches

    def __upsert_mssql(self, schema: str, table: str, cols: [], row: [], conflict_cols: []):
        table = f"{schema}.{table}" if schema else table
        target_cols = ', '.join(cols)
        source_cols = ', '.join([f"source.{col}" for col in cols])
        update_clause = ', '.join([f"target.{col} = source.{col}" for col in cols if col not in conflict_cols])
        conflict_cdt = ' AND '.join([f"target.{col} = source.{col}" for col in conflict_cols])
        query = f"""
        MERGE INTO {table} AS target
        USING (VALUES ({', '.join(['%s'] * len(cols))})) AS source ({', '.join(cols)})
        ON {conflict_cdt}
        WHEN MATCHED THEN
            UPDATE SET {update_clause}
        WHEN NOT MATCHED THEN
            INSERT ({target_cols})
            VALUES ({source_cols});"""
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeur(s) insérée(s) ou mise(s) à jour dans la table {table}.")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur lors de l'upsert dans la table {table}.")
            return "ERROR", str(e), row

    def __upsert_pg(self, schema: str, table: str, cols: [], row: [], conflict_cols: []):
        table = f"{schema}.{table}" if schema else table
        placeholders = ', '.join(["%s"] * len(cols))
        conflict_clause = ', '.join(conflict_cols)
        update_clause = ', '.join([f"{col} = EXCLUDED.{col}" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table} ({', '.join(cols)})
        VALUES ({placeholders})
        ON CONFLICT ({conflict_clause}) DO UPDATE
        SET {update_clause};
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeurs insérée(s) ou mise(s) à jour dans la table {table}.")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur lors de l'upsert dans la table {table}: {e}")
            return "ERROR", str(e), row

    def __upsert_mysql(self, schema: str, table: str, cols: [], row: [], conflict_cols: []):
        table_full = f"{schema}.{table}" if schema else table
        columns = ', '.join([f"'{col}'" for col in cols])
        placeholders = ', '.join(['%s'] * len(cols))
        update_clause = ', '.join([f"'{col}' = VALUES('{col}')" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table_full} ({columns})
        VALUES ({placeholders})
        ON DUPLICATE KEY UPDATE {update_clause};
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeurs insérée(s) ou mise(s) à jour dans la table {table_full}.")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur lors de l'upsert dans la table {table}.")
            return "ERROR", str(e), row

    @timer
    def upsert(self, schema: str, table: str, cols: [], row: [], conflict_cols: []):
        """
        Réalise une opération upsert sur la base
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param row: Liste des valeurs à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et la ligne en cas d'erreur
        """
        if self.__techno == "mssql":
            result, errors, failed_batches = self.__upsert_mssql(schema, table, cols, row, conflict_cols)
        elif self.__techno == "postgresql":
            result, errors, failed_batches = self.__upsert_pg(schema, table, cols, row, conflict_cols)
        elif self.__techno == "mysql":
            result, errors, failed_batches = self.__upsert_mysql(schema, table, cols, row, conflict_cols)
        else:
            msg = f"La connexion {self.__techno} n'est pas supportée. Tentez avec 'mssql', 'postgresql' ou 'mysql'"
            self.logger.error(msg)
            raise ValueError(msg)
        return result, errors, failed_batches

    def __upsertMany_mssql(self, schema: str, table: str, cols: [], rows: [[]], conflict_cols: []):
        failed_batches = []
        errors = []
        total_inserted = 0
        table = f"{schema}.{table}" if schema else table
        target_cols = ', '.join(cols)
        source_cols = ', '.join([f"source.{col}" for col in cols])
        update_clause = ', '.join([f"target.{col} = source.{col}" for col in cols if cols not in conflict_cols])
        conflict_cdt = ' AND '.join([f"target.{col} = source.{col}" for col in conflict_cols])
        query = f"""
        MERGE INTO {table} AS target
        USING (VALUES ({', '.join(['%s'] * len(cols))})) AS source ({', '.join(cols)})
        ON {conflict_cdt}
        WHEN MATCHED THEN
            UPDATE SET {update_clause}
        WHEN NOT MATCHED THEN
            INSERT ({target_cols})
            VALUES ({source_cols});"""
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) ou mise(s) à jour dans la table {table}. "
                            f"Total inséré ou mis à jour: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'upsert du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
                return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'upsert des données: {e}")
            return "ERROR", str(e), rows

    def __upsertMany_pg(self, schema: str, table: str, cols: [], rows: [[]], conflict_cols: []):
        failed_batches = []
        errors = []
        total_inserted = 0
        table = f"{schema}.{table}" if schema else table
        placeholders = ', '.join(["%s"] * len(cols))
        conflict_clause = ', '.join(conflict_cols)
        update_clause = ', '.join([f"{col} = EXCLUDED.{col}" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table} ({', '.join(cols)})
        VALUES ({placeholders})
        ON CONFLICT ({conflict_clause}) DO UPDATE
        SET {update_clause};
        """
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérées ou mises à jour dans la table {table}. "
                            f"Total inséré ou mis à jour: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'upsert du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
                return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'upsert des données: {e}")
            return "ERROR", str(e), rows

    def __upsertMany_mysql(self, schema: str, table: str, cols: [], rows: [[]], conflict_cols: []):
        failed_batches = []
        errors = []
        total_inserted = 0
        table_full = f"{schema}.{table}" if schema else table
        columns = ', '.join([f"'{col}'" for col in cols])
        placeholders = ', '.join(['%s'] * len(cols))
        update_clause = ', '.join([f"'{col}' = VALUES('{col}')" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table_full} ({columns})
        VALUES ({placeholders})
        ON DUPLICATE KEY UPDATE {update_clause};
        """
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) ou mise(s) à jour dans la table {table}. "
                            f"Total inséré ou mis à jour: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'upsert du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
            return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'upsert des données: {e}")
            return "ERROR", str(e), rows

    @timer
    def upsertMany(self, schema: str, table: str, cols: [], rows: [[]], conflict_cols: []):
        """
        Réalise une opération upsert par batch sur la base
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param rows: Liste des lignes à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et le batch en cas d'erreur
        """
        if self.__techno == "mssql":
            result, errors, failed_batches = self.__upsertMany_mssql(schema, table, cols, rows, conflict_cols)
        elif self.__techno == "postgresql":
            result, errors, failed_batches = self.__upsertMany_pg(schema, table, cols, rows, conflict_cols)
        elif self.__techno == "mysql":
            result, errors, failed_batches = self.__upsertMany_mysql(schema, table, cols, rows, conflict_cols)
        else:
            msg = f"La connexion {self.__techno} n'est pas supportée. Tentez avec 'mssql', 'postgresql' ou 'mysql'"
            self.logger.error(msg)
            raise ValueError(msg)
        return result, errors, failed_batches

    def __create_temp_table(self, schema: str, table: str, temp_table: str, cols: []):
        query = f"""
        SELECT COLUMN_NAME, DATA_TYPE, CHARACTER_MAXIMUM_LENGTH
        FROM INFORMATION_SCHEMA.COLUMNS
        WHERE TABLE_SCHEMA = '{schema}' AND TABLE_NAME = '{table}'
        """
        self.logger.info(f"Récupération des colonnes depuis {schema}.{table}")
        column_definitions = list(self.sqlQuery(query))[0]
        col_def = []
        for col in cols:
            col_info = next((c for c in column_definitions if c[0] == col), None)
            if col_info:
                col_name, data_type, char_length = col_info
                if char_length:
                    col_def.append(f"{col_name} {data_type.upper()}({char_length})")
                else:
                    col_def.append(f"{col_name} {data_type.upper()}")
            else:
                raise ValueError(f"Column {col} not found in {schema}.{table}")
        col_def_sql = ', '.join(col_def)
        self.sqlExec(f"DROP TABLE IF EXISTS {schema}.{temp_table}")
        self.sqlExec(f"CREATE TABLE {schema}.{temp_table} ({col_def_sql})")

    def __upsertBulk_mssql(self, schema: str, table: str, cols: [], rows, conflict_cols):
        temp_table = f"{table}_temp"
        table_full = f"{schema}.{table}" if schema else table
        conflict_cdt = ' AND '.join([f"target.{col} = source.{col}" for col in conflict_cols])
        update_clause = ', '.join([f"target.{col} = source.{col}" for col in cols if col not in conflict_cols])
        target_cols = ', '.join(cols)
        source_cols = ', '.join([f"source.{col}" for col in cols])
        query = f"""
        MERGE INTO {table_full} AS target
        USING {temp_table} AS source
        ON {conflict_cdt}
        WHEN MATCHED THEN
            UPDATE SET {update_clause}
        WHEN NOT MATCHED THEN
            INSERT ({target_cols})
            VALUES ({source_cols});
        """
        try:
            self.__create_temp_table(schema, table, temp_table, cols)
            result, errors, failed_batches = self.insertBulk(schema, temp_table, cols, rows)
            if result == "ERROR":
                return result, errors, failed_batches
            self.sqlExec(query)
            self.logger.info(f"{len(rows)} ligne(s) insérée(s) ou mise(s) à jour dans la tables {table}.")
            return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'UPSERT: {e}")
            return "ERROR", str(e), rows
        finally:
            self.sqlExec(f"DROP TABLE IF EXISTS {schema}.{temp_table}")
            self.logger.info(f"Table temporaire supprimée.")

    def __upsertBulk_pg(self, schema: str, table: str, cols: [], rows, conflict_cols: []):
        temp_table = f"{table}_temp"
        table_full, temp_table_full = (f"{schema}.{table}", f"{schema}.{temp_table}") if schema else (table, temp_table)
        conflict_clause = ', '.join(conflict_cols)
        update_clause = ', '.join([f"{col} = EXCLUDED.{col}" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table_full} ({', '.join(cols)})
        SELECT {', '.join(cols)} FROM {temp_table_full}
        ON CONFLICT ({conflict_clause}) DO UPDATE
        SET {update_clause};
        """
        try:
            self.sqlExec(f"CREATE TEMP TABLE {temp_table_full} (LIKE {table_full} INCLUDING DEFAULTS)")
            result, errors, failed_batches = self.insertBulk(schema, temp_table, cols, rows)
            if result == "ERROR":
                return result, errors, failed_batches
            self.sqlExec(query)
            self.logger.info(f"{len(rows)} lignes insérée(s) ou mise(s) à jour dans la table {table}.")
            return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'UPSERT : {e}")
            return "ERROR", str(e), rows
        finally:
            self.sqlExec(f"DROP TABLE IF EXISTS {temp_table_full}")
            self.logger.info(f"Table temporaire supprimée.")

    def __upsertBulk_mysql(self, schema: str, table: str, cols: [], rows, conflict_cols: []):
        temp_table = f"{table}_temp"
        table_full, temp_table_full = (f"{schema}.{table}", f"{schema}.{temp_table}") if schema else (table, temp_table)
        columns = ', '.join(cols)
        update_clause = ', '.join([f"{col} = VALUES({col})" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table_full} ({columns})
        SELECT {columns} FROM {temp_table_full}
        ON DUPLICATE KEY UPDATE {update_clause};
        """
        try:
            self.sqlExec(f"CREATE TEMPORARY TABLE {temp_table_full} LIKE {table_full}")
            self.logger.info(f"Table temporaire {temp_table_full} créée.")
            result, errors, failed_batches = self.insertBulk(schema, temp_table, cols, rows)
            if result == "ERROR":
                return result, errors, failed_batches
            self.sqlExec(query)
            self.logger.info(f"{len(rows)} lignes insérée(s) ou mise(s) à jour dans la table {table}.")
            return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'UPSERT : {e}")
            return "ERROR", str(e), rows
        finally:
            self.sqlExec(f"DROP TEMPORARY TABLE IF EXISTS {temp_table_full}")
            self.logger.info(f"Table temporaire {temp_table_full} supprimée.")

    @timer
    def upsertBulk(self, schema: str, table: str, cols: [], rows, conflict_cols: []):
        """
        Réalise une opération upsert en bulk par batch sur la base en lots et avec l'aide d'une table temporaire
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param rows: Liste des lignes à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et le batch en cas d'erreur
        """
        if self.__techno == "mssql":
            result, errors, failed_batches = self.__upsertBulk_mssql(schema, table, cols, rows, conflict_cols)
        elif self.__techno == "postgresql":
            result, errors, failed_batches = self.__upsertBulk_pg(schema, table, cols, rows, conflict_cols)
        elif self.__techno == "mysql":
            result, errors, failed_batches = self.__upsertBulk_mysql(schema, table, cols, rows, conflict_cols)
        else:
            msg = f"La connexion {self.__techno} n'est pas supportée. Tentez avec 'mssql', 'postgresql' ou 'mysql'"
            self.logger.error(msg)
            raise ValueError(msg)
        return result, errors, failed_batches


from abc import ABC, abstractmethod

class ReaderHelper(ABC):
    """
    Interface pour aider à paramétrer un Reader.
    """
    def __init__(self, connection):
        """
        Initialise le ReaderHelper avec une connexion donnée.
        :param connection: Instance de la classe connection utilisée pour accèder aux données.
        """
        self.connection = connection

    @abstractmethod
    def get_columns(self, source: str):
        """
        Récupère les colonnes disponibles dans la source (table, fichier...).
        :param source: Requête SQL ou chemin du fichier.
        :return: Liste des colonnes (nom et type).
        """
        pass

    @abstractmethod
    def guess_read(self, source: str):
        """
        Devine les paramètres nécéssaires pour lire les données.
        :param source: Requête SQL ou chemin du fichier.
        :return: Paramètres devinés (structure, types...).
        """
        pass


import requests

_timer_enabled = False  # Variable globale pour déterminer si le timer est actif ou non

def set_timer(state: bool):
    """
    Active ou désactive le timer
    :param state: détermine l'état du timer
    """
    global _timer_enabled
    _timer_enabled = state

def get_timer() -> bool:
    """
    récupère l'état du timer
    :return: l'état du timer
    """
    return _timer_enabled

def get_public_ip(timeout: int = 5) -> str:
    """
    Récupère l'adresse IP publique de la machine
    :param timeout: Délai max de la requête http en secondes
    """
    urls = [
        "https://api.ipify.org",
        "https://ifconfig.me/ip",
        "https://checkup.amazonaws.com"
    ]
    for url in urls:
        try:
            response = requests.get(url, timeout=timeout)
            if response.status_code == 200:
                return response.text.strip()
        except requests.RequestException:
            continue
    raise RuntimeError("Impossible de récupérer l'adresse IP publique")


from abc import ABC, abstractmethod

class data_factory(ABC):
    """Classe abstraite qui pose un modèle de connexion ads"""
    @abstractmethod
    def connect(self):
        """Lance la connexion avec les identifiants passés à l'initialisation de la classe"""
        pass

    @abstractmethod
    def insert(self, schema: str, table: str, cols: [], row: []):
        """
        Insère des données dans la base de données, nécessite une connexion active
        :param schema: nom du schéma dans lequel insérer
        :param table: nom de la table dans laquelle insérer
        :param cols: liste des colonnes dans lesquelles insérer
        :param row: liste des valeurs à insérer
        """
        pass

    @abstractmethod
    def insertBulk(self, schema: str, table: str, cols: [], rows):
        """
        Similaire à insert classique, mais insère par batch de taille 'batch_size' définie
        :param schema: nom du schéma dans lequel insérer
        :param table: nom de la table dans laquelle insérer
        :param cols: liste des colonnes dans lesquelles insérer
        :param rows: liste des lignes à insérer
        """
        pass

    @abstractmethod
    def sqlQuery(self, query: str, return_columns: bool = False):
        """
        Lit la base de données avec la requête query
        :param return_columns: le premier élément retourné sera la liste des colonnes
        :param query: la requête
        :return: un générateur de liste de tuples
        """
        pass

    @abstractmethod
    def sqlExec(self, query):
        """
        Execute une requête sur la base de données
        :param query: la requête
        """
        pass


from typing import Callable, List, Dict

class Pipeline:
    """
    Pipeline qui lie un Reader à un Writer et exécute un flux de données
    """
    def __init__(self, reader, writer, logger=None):
        """
        Initialise le Pipeline avec un Reader et un Writer
        :param reader: Instance d'un Reader pour lire les données
        :param writer: Instance d'un Writer pour écrire les données
        :param logger: Logger pour suivre l'exécution du pipeline
        """
        self.reader = reader
        self.writer = writer
        self.logger = logger
        self.column_mapping = {}
        self.calculated_columns = []
        self.join_sources = []

    def set_mapping(self, column_mapping: Dict[str, str]):
        """
        Définit le mapping entre les colonnes sources et les colonnes cibles
        :param column_mapping: Dictionnaire {colonne_source: colonne_cible}.
        """
        self.column_mapping = column_mapping

    def add_calculated_column(self, column_name: str, function: Callable):
        """
        Ajoute une colonne calculée au flux.
        :param column_name: Nom de la nouvelle colonne
        :param function: Fonction qui calcule la colonne (prend une ligne en entrée et retourne une valeur)
        """
        self.calculated_columns.append((column_name, function))

    def add_join(self, join_reader, join_key_source: str, join_key_target: str, join_type: str = "inner"):
        """
        Ajoute une jointure avec une autre source de données.
        :param join_reader: Instance d'un Reader pour la source à joindre.
        :param join_key_source: Clé de jointure dans les données sources.
        :param join_key_target: Clé de jointure dans les données cibles.
        :param join_type: Type de jointure (inner, left, right, full).
        """
        self.join_sources.append({
            "reader": join_reader,
            "source_key": join_key_source,
            "target_key": join_key_target,
            "type": join_type,
        })

    def _apply_mapping(self, row: Dict):
        """
        Applique le mapping des colonnes sur une ligne de données.
        :param row: Ligne de données source.
        :return: Ligne de données avec les colonnes mappées.
        """
        if not self.column_mapping:
            return row
        return {self.column_mapping.get(key, key): value for key, value in row.items()}

    def _apply_calculations(self, row: Dict):
        """
        Applique les colonnes calculées sur une ligne de données.
        :param row: Ligne de données.
        :return: Ligne de données enrichie avec les colonnes calculées
        """
        for column_name, function in self.calculated_columns:
            row[column_name] = function(row)
        return row

    def _join_data(self, rows: List[Dict]):
        """
        Applique les jointures entre les données sources et les autres sources.
        :param rows: Ligne de données sources.
        :return: Ligne de données jointes.
        """
        for join in self.join_sources:
            join_reader = join["reader"]
            source_key = join["source_key"]
            target_key = join["target_key"]
            join_type = join["type"]
            joined_rows = list(join_reader.read())
            join_index = {row[target_key]: row for row in joined_rows}
            for row in rows:
                join_value = row.get(source_key)
                if join_value in join_index:
                    row.update(join_index[join_value])
                elif join_type == "left":
                    continue
                elif join_type == "inner":
                    rows.remove(row)
                elif join_type == "right":
                    raise NotImplementedError("Right join non supportée dans cet exemple.")
        return rows

    def execute(self, source_query_or_path: str, mode: str = "insert", batch_size: int = None):
        """
        Exécute le flux de données entre le Reader et le Writer.
        :param source_query_or_path: Requête SQL ou chemin de fichier pour le Reader.
        :param mode: Mode d'écriture (insert, update, etc.)
        :param batch_size: Taille des lots à lire et à écrire
        """
        if self.logger:
            self.logger.info("Début de l'exécution du pipeline.")
        for batch in self.reader.read(source_query_or_path, batch_size=batch_size):
            if isinstance(batch, list):
                batch = [dict(row) for row in batch]
            batch = [self._apply_mapping(row) for row in batch]
            batch = [self._apply_calculations(row) for row in batch]
            batch = self._join_data(batch)
            self.writer.write(batch, mode=mode, batch_size=batch_size)
        if self.logger:
            self.logger.info("Exécution du pipeline terminée avec succès.")


import os
import timeit
import chardet
import smbclient
import unicodedata
from typing import List
from .timer import timer, get_timer
from .dataFactory import data_factory

class FileHandler:
    def __init__(self, logger, smb_config: dict=None, batch_size: int=4_096):
        """Initialise le gestionnaire de fichiers avec le chemin du fichier"""
        self.logger = logger
        self.smb_config = smb_config
        self.batch_size = batch_size
        if smb_config:
            self._setup_smb_connection()

    @timer
    def _setup_smb_connection(self):
        """
        Configure la connection SMB
        """
        try:
            smbclient.register_session(
                self.smb_config.get("server"),
                username=self.smb_config.get("username"),
                password=self.smb_config.get("password")
            )
            self.logger.info("Connexion SMB établie.")
        except Exception as e:
            self.logger.error(f"Échec de la connexion SMB : {e}")
            raise

    def read_file(self, file_path: str, mode: str='rb', encoding: str=None):
        """Lit un fichier local ou SMB en batchs"""
        if self.smb_config and file_path.startswith("//"):
            return self._read_smb_file(file_path, mode, encoding)
        return self._read_local_file(file_path, mode, encoding)

    def _read_smb_file(self, file_path, mode, encoding):
        """Lit un fichier sur un partage SMB en batchs"""
        timer_start = timeit.default_timer()
        try:
            with smbclient.open_file(file_path, mode=mode, encoding=encoding) as file:
                while chunk := file.read(self.batch_size):
                    yield chunk
                self.logger.info(f"Lecture réussie du fichier : {file_path}")
            if get_timer():
                elapsed_time = timeit.default_timer() - timer_start
                self.logger.info(f"Temps d'exécution de sqlQuery: {elapsed_time:.4f} secondes")
        except Exception as e:
            self.logger.error(f"Erreur de lecture du fichier SMB : {e}")
            raise

    def _read_local_file(self, file_path, mode, encoding):
        """Lit un fichier en local en batchs"""
        timer_start = timeit.default_timer()
        try:
            with open(file_path, mode=mode, encoding=encoding) as file:
                while chunk := file.read(self.batch_size):
                    yield chunk
                self.logger.info(f"Lecture réussie du fichier: {file_path}")
            if get_timer():
                elapsed_time = timeit.default_timer() - timer_start
                self.logger.info(f"Temps d'exécution de sqlQuery: {elapsed_time:.4f} secondes")
        except Exception as e:
            self.logger.error(f"Erreur de lecture du fichier en local: {e}")
            raise

    @timer
    def write_file(self, file_path: str, content, mode: str='wb', clean: bool=False):
        """Écrit du contenu dans un fichier local ou SMB en batchs"""
        if clean:
            self.logger.info(f"Nettoyage du fichier {file_path}.")
        if self.smb_config and file_path.startswith("//"):
            return self._write_smb_file(file_path, content, mode, clean)
        return self._write_local_file(file_path, content, mode, clean)

    def _write_smb_file(self, file_path, content, mode, clean: bool=False):
        """Écrit du contenu dans un fichier via partage SMB en batchs"""
        try:
            encoding = "utf-8" if "b" not in mode else None
            with smbclient.open_file(file_path, mode=mode, encoding=encoding) as file:
                for chunk in content:
                    if clean:
                        chunk = self.clean_text(chunk)
                    for i in range(0, len(chunk), self.batch_size):
                        file.write(chunk[i:i + self.batch_size])
                self.logger.info(f"Succès: contenu écrit dans '{file_path}' (SMB).")
        except Exception as e:
            self.logger.error(f"Erreur lors de l'écriture du fichier '{file_path}' via SMB: {e}")
            raise

    def _write_local_file(self, file_path, content, mode, clean: bool=False):
        """Écrit du contenu dans un fichier local en batchs"""
        try:
            encoding = "utf-8-sig" if "b" not in mode else None
            with open(file_path, mode=mode, encoding=encoding) as file:
                for chunk in content:
                    if clean:
                        chunk = self.clean_text(chunk)
                    for i in range(0, len(chunk), self.batch_size):
                        file.write(chunk[i:i + self.batch_size])
                self.logger.info(f"Succès: Contenu écrit dans '{file_path}' (local).")
        except Exception as e:
            self.logger.error(f"Erreur lors de l'écriture du fichier '{file_path}' en local: {e}")
            raise

    def list_dir(self, dir_path: str):
        """Liste le contenu d'un dossier en local ou via SMB"""
        try:
            if self.smb_config and dir_path.startswith("//"):
                return smbclient.listdir(dir_path)
            return os.listdir(dir_path)
        except Exception as e:
            self.logger.error(f"Erreur lors de la lecture du dossier {dir_path}: {e}")
            return []

    def file_exists(self, file_path: str):
        """Vérifie si le fichier existe (local ou SMB)"""
        if self.smb_config and file_path.startswith("//"):
            return self._smb_file_exists(file_path)
        return self._local_file_exists(file_path)

    def _smb_file_exists(self, file_path):
        """Vérifie si un fichier SMB existe"""
        try:
            exists = smbclient.path.exists(file_path) and smbclient.path.isfile(file_path)
            if exists:
                self.logger.info(f"Le fichier existe: {file_path}")
            else:
                self.logger.warning(f"Le fichier n'existe pas: {file_path}")
            return exists
        except Exception as e:
            self.logger.error(f"Erreur lors de la vérification d'existence du fichier {file_path}: {e}")

    def _local_file_exists(self, file_path):
        """Vérifie si un fichier en local existe"""
        exists = os.path.exists(file_path) and os.path.isfile(file_path)
        if exists:
            self.logger.info(f"Le fichier existe: {file_path}")
        else:
            self.logger.warning(f"Le fichier n'existe pas: {file_path}")
        return exists

    def directory_exists(self, dir_path: str):
        """Vérifie si le fichier existe (local ou SMB)"""
        if self.smb_config and dir_path.startswith("//"):
            return self._smb_directory_exists(dir_path)
        return self._local_directory_exists(dir_path)

    def _smb_directory_exists(self, dir_path: str):
        try:
            exists = smbclient.path.exists(dir_path) and smbclient.path.isdir(dir_path)
            if exists:
                self.logger.info(f"Le dossier existe: {dir_path}")
            else:
                self.logger.warning(f"Le dossier n'existe pas: {dir_path}")
            return exists
        except Exception as e:
            self.logger.error(f"Erreur lors de la vérification d'existence du dossier {dir_path}: {e}")

    def _local_directory_exists(self, dir_path):
        exists = os.path.exists(dir_path) and os.path.isdir(dir_path)
        if exists:
            self.logger.info(f"Le dossier existe: {dir_path}")
        else:
            self.logger.warning(f"Le dossier n'existe pas: {dir_path}")
        return exists

    def remove_file(self, file_path: str):
        """Supprime un fichier (local ou SMB)"""
        try:
            if self.smb_config and file_path.startswith("//"):
                smbclient.remove(file_path)
            else:
                os.remove(file_path)
            self.logger.info("Fichier supprimé avec succès.")
        except Exception as e:
            self.logger.error(f"Erreur lors de la suppression du fichier '{file_path}': {e}")
            raise

    def create_empty_file(self, file_path: str, mode: str='wb'):
        """Crée un fichier vide (local ou SMB)"""
        try:
            if self.smb_config and file_path.startswith("//"):
                with smbclient.open_file(file_path, mode=mode) as file:
                    pass
                self.logger.info(f"Fichier vide créé sur SMB: {file_path}")
            else:
                with open(file_path, mode=mode) as file:
                    pass
                self.logger.info(f"Fichier vide créé localement: {file_path}")
        except Exception as e:
            self.logger.error(f"Erreur lors de la création du fichier vide '{file_path}': {e}")
            raise

    def clean_text(self, text):
        if isinstance(text, bytes):
            encoding = chardet.detect(text)["encoding"]
            text = text.decode(encoding or "utf-8", errors='ignore')
        text = unicodedata.normalize('NFKD', text)
        text = ''.join([c for c in text if not unicodedata.combining(c)])
        text = ''.join([f'&#{ord(c)};' if ord(c) > 127 else c for c in text])
        return text.encode("utf-8")

    def _get_dir_size(self, path='.'):
        """Calcule récursivement la taille d'un répertoire"""
        total = 0
        with os.scandir(path) as it:
            for entry in it:
                try:
                    if entry.is_file():
                        total += entry.stat().st_size
                    elif entry.is_dir():
                        total += self._get_dir_size(entry.path)
                except Exception as e:
                    self.logger.error(f"Erreur sur le fichier {entry.name}: {e}")
                    raise
        return total

    def _convertir_octets(self, taille_octets):
        """Convertit une taille en octets dans une unité lisible"""
        for unite in ['octets', 'Ko', 'Mo', 'Go', 'To']:
            if taille_octets < 1024.0:
                return round(taille_octets, 2), unite
            taille_octets /= 1024.0
        return round(taille_octets, 2), 'To'

    def _ensure_table_exists(self, db: data_factory, schema: str, table: str):
        """Crée la table cible si elle n'existe pas encore."""
        from .dbMysql import dbMysql
        from .dbMssql import dbMssql
        from .dbPgsql import dbPgsql

        cols = ["tenantname", "date", "taille", "unite", "fichier"]
        cols_def = ["VARCHAR(255)", "TIMESTAMP", "FLOAT", "VARCHAR(10)", "VARCHAR(255)"]
        if isinstance(db, dbMssql):
            cols_def[1] = "DATETIME2"
        columns = ", ".join(f"{col} {typ}" for col, typ in zip(cols, cols_def))
        full_table = f"{schema}.{table}" if schema else table
        if isinstance(db, (dbMysql, dbPgsql)):
            query = f"CREATE TABLE IF NOT EXISTS {full_table} ({columns});"
        elif isinstance(db, dbMssql):
            query = f"""
            IF OBJECT_ID('{full_table}', 'U') IS NULL
            CREATE TABLE {full_table} ({columns});
            """
        else:
            raise ValueError(f"Type de base de données non pris en charge: {type(db).__name__}")
        db.sqlExec(query)

    def _collect_files(self, base_path: str, filter: str):
        """Retourne la liste complète des fichiers correspondant au filtre."""
        import os
        return [
            os.path.join(root, f)
            for root, _, files in os.walk(base_path)
            for f in files
            if filter in f
        ]

    def _process_files(self, db: data_factory, schema: str, table: str, fichiers: List[str], tenant_name: str, date: str, base_path: str):
        """Traite les fichiers, calcule les tailles et insère les résultats dans la base."""
        total_size = 0
        bulk_data = []
        cols = ["tenantname", "date", "taille", "unite", "fichier"]

        for fichier in fichiers:
            try:
                size = os.path.getsize(fichier)
                total_size += size
                value, unit = self._convertir_octets(size)
                bulk_data.append([tenant_name, date, str(value), unit, fichier])
            except Exception as e:
                self.logger.error(f"Erreur sur le fichier {fichier}: {e}")

        if bulk_data:
            db.insertBulk(schema, table, cols, bulk_data)

        total_value, total_unit = self._convertir_octets(total_size)
        total_data = [f"{tenant_name} TOTAL", date, str(total_value), total_unit, f"TOTAL {base_path}"]
        db.insert(schema, table, cols, total_data)

        self.logger.info(f"Taille totale: {total_value} {total_unit}")

    @timer
    def disk_check(self, db: data_factory, schema: str, table: str, base_path: str, tenant_name: str = 'DEFAULT', filter: str = 'python-'):
        """
        Calcule la taille des fichiers correspondant à un filtre dans un répertoire
        et enregistre les résultats dans la table {schema}.{table}.
        :param db: Instance de data_factory (dbMssql, dbPgsql, dbMysql)
        :param schema: Nom du schéma
        :param table: Nom de la table cible
        :param base_path: Répertoire racine à scanner
        :param tenant_name: Nom du tenant à enregistrer
        :param filter: Mot-clé à rechercher dans les fichiers
        """
        from datetime import date

        self.logger.info(f"Lancement du disk_check à partir de {base_path}")
        db.connect()
        today = str(date.today())

        self._ensure_table_exists(db, schema, table)
        fichiers = self._collect_files(base_path, filter)
        if not fichiers:
            self.logger.warning(f"Aucun fichier contenant '{filter}' trouvé dans {base_path}.")
            return
        self._process_files(db, schema, table, fichiers, tenant_name, today, base_path)

    def wait_for_file(self, path: str, filename: str, retry: int = 20):
        """
        Attend qu'un fichier spécifique soit disponible dans un répertoire donné (local ou SMB),
        avec un nombre limité de tentatives
        :param path: Chemin du répertoire où chercher le fichier
        :param filename: Nom du fichier à attendre
        :param retry: Nombre maximal de tentatives
        :return: True si le fichier est trouvé dans le nombre de tentatives sinon False
        """
        import time

        full_path = os.path.join(path, filename)
        self.logger.info(f"Attente du fichier '{filename}' dans le répertoire '{path}'.")

        try:
            if self.smb_config and path.startswith("//"):
                exists = smbclient.path.exists(path)
            else:
                exists = os.path.exists(path)
        except Exception as e:
            self.logger.error(f"Erreur lors de la vérification du chemin '{path}': {e}")
            raise
        if not exists:
            msg = f"Le chemin spécifié n'existe pas: {path}"
            self.logger.error(msg)
            raise ValueError(msg)
        i = 1
        while i <= retry:
            try:
                if self.smb_config and full_path.startswith("//"):
                    found = smbclient.path.isfile(full_path)
                else:
                    found = os.path.isfile(full_path)
            except Exception as e:
                self.logger.error(f"Erreur lors de la vérification du fichier '{full_path}': {e}")
                raise
            if found:
                self.logger.info(f"Fichier '{filename}' trouvé après {i} tentative(s).")
                return True
            time.sleep(1)
            i += 1
        self.logger.warning(f"Fichier '{filename}' introuvable après {retry} tentative(s).")
        return False


import re
import email
import chardet
import imaplib
import requests
import mimetypes
from email import policy

class mail:
    def __init__(self, configuration: dict, logger, auth_mode: str):
        """
        Instancie la classe mail
        :param configuration: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion imap
        :param logger: Un logger ads qui va gérer les logs des actions de la classe
        :param auth_mode: Le mode d'authentification 'oauth2' ou 'credentials'
        """
        self.logger = logger
        self.__url=configuration.get("url")
        self.__clientId=configuration.get("clientId")
        self.__clientSecret=configuration.get("clientSecret")
        self.__email=configuration.get("email")
        self.__emailFrom=configuration.get("emailFrom")
        self.__subject=configuration.get("subject")
        self.__scope=configuration.get("scope")
        self.__server=configuration.get("server")
        self.__login=configuration.get("login")
        self.__password=configuration.get("password")
        self.__token=None
        self.__connection=None
        if auth_mode == 'oauth2':
            self.__connect_with_token()
        elif auth_mode == 'credentials':
            self.__connect_with_credentials()
        else:
            raise ValueError (f"Le mode d'authentification {auth_mode} n'est pas reconnu. Utiliser 'oauth2' ou 'credentials'")

    def __connect_with_credentials(self):
        """Se connecte à un server imap et récupère le token"""
        if self.logger: self.logger.info("Tentative de connexion avec le serveur mail.")
        try:
            imap_conn = imaplib.IMAP4_SSL(self.__server)
            # Authentification avec les identifiants utilisateur
            imap_conn.login(self.__login, self.__password)
            self.__connection = imap_conn
            if self.logger: self.logger.info("Connexion établie.")
        except Exception as e:
            if self.logger: self.logger.error(f"La tentative de connexion a échoué : {str(e)}")
            raise

    def __connect_with_token(self):
        """Se connecte à un server imap et récupère le token"""
        if self.logger: self.logger.info("Tentative de connexion avec le server mail.")
        payload = {
            'grant_type': 'client_credentials',
            'client_id': self.__clientId,
            'client_secret': self.__clientSecret,
            'scope': self.__scope
        }
        response = requests.post(self.__url, data=payload)
        if response.status_code==200:
            if self.logger: self.logger.info("Connexion établie")
            if self.logger: self.logger.debug("Tentative de récupération du token.")
            token_data = response.json()
            if "access_token" in token_data:
                if self.logger: self.logger.debug("Token récupéré avec succès !")
                self.__token = token_data["access_token"]
                imap_conn = imaplib.IMAP4_SSL(self.__server)
                auth_string = f"user={self.__email}\1auth=Bearer {self.__token}\1\1"
                imap_conn.authenticate("XOAUTH2", lambda x: auth_string.encode("utf-8"))
                self.__connection = imap_conn
            else:
                self.logger.error(f"Le token n'a pas pu être récupéré {response.reason}.")
                raise

        else:
            self.logger.error(f"La tentative de connexion a echoué {response.reason}.")
            raise

    def mark_as_seen(self, id):
        imap_conn = self.__connection
        if self.logger: self.logger.info(f"Changement du flag du mail {id} à 'Seen'.")
        try:
            imap_conn.store(id, '+FLAGS', '\\Seen')
        except Exception as e:
            if self.logger: self.logger.info(f"Le changement de flag a échoué: {str(e)}.")
            raise

    def mark_as_un_seen(self, id):
        imap_conn = self.__connection
        if self.logger: self.logger.info(f"Changement du flag du mail {id} à 'UnSeen'.")
        try:
            imap_conn.store(id, '-FLAGS', '\\UnSeen')
        except Exception as e:
            if self.logger: self.logger.info(f"Le changement de flag a échoué: {str(e)}.")
            raise

    def delete(self,id):
        imap_conn = self.__connection
        if self.logger: self.logger.info(f"Suppression du mail {id}.")
        try:
            imap_conn.store(id, '+FLAGS', '\\Deleted')
        except Exception as e:
            if self.logger: self.logger.info(f"La suppression a échoué: {str(e)}.")
            raise

    def get_email(self, mailbox='INBOX', readonly=False, emailFrom=None, emailSubject=None, flag="UnSeen", charset=None):
        """
        Récupère les mails selon les paramètres renseignés
        :mailbox string: Contient le dossier de lecture (ex : INBOX)
        :readonly bool: définit si le mail est modifiable. Si False, alors le mail sera automatiqumeent marqué comme lu.
        :emailFrom string: Permet de filtrer sur le destinataire du mail
        :emailSubject string: Permet de filtrer sur l'objet du mail
        :flag string: Permet de filtrer sur le flag ;
            "ALL",
            "SEEN",
            "UNSEEN",
            "FLAGGED",
            "UNFLAGGED",
            "ANSWERED",
            "UNANSWERED",
            "DELETED",
            "UNDELETED",
            "DRAFT"
        :charset string: Permet de définir le charset ex : UTF-8, US-ASCII
        """
        matching_emails=[]
        imap_search_criteria = [
            "ALL",
            "SEEN",
            "UNSEEN",
            "FLAGGED",
            "UNFLAGGED",
            "ANSWERED",
            "UNANSWERED",
            "DELETED",
            "UNDELETED",
            "DRAFT"
        ]
        if flag.upper() not in imap_search_criteria:
            msg = ('Flag incorrect :  ["ALL", "SEEN", "UNSEEN", "FLAGGED", "UNFLAGGED", "ANSWERED", '
                   '"UNANSWERED", "DELETED", "UNDELETED", "DRAFT"]')
            if self.logger: self.logger.error(msg)
            raise ValueError(msg)
        if self.logger: self.logger.info("Tentative de récupération des mails.")
        if not self.__connection:
            if self.logger: self.logger.error("Une connexion doit être établie avant de réaliser cette opération.")
            raise
        imap_conn=self.__connection
        imap_conn.select(mailbox.upper(), readonly)
        if emailFrom :
            status, email_ids = imap_conn.search(charset, f'{flag} FROM "{emailFrom}"')
        else:
            status, email_ids = imap_conn.search(charset, str(flag))
        if status != "OK":
            if self.logger: self.logger.error(
                f"Erreur de la récupération des emails, {status, email_ids[0]}")
        if not email_ids[0]:
            if self.logger: self.logger.info(f"Aucun email n'a été trouvé.")
            return []
        email_ids = email_ids[0].split()
        if emailSubject:
            for e_id in email_ids:
                status, msg_data = imap_conn.fetch(e_id, '(RFC822)')
                for response_part in msg_data:
                    if isinstance(response_part, tuple):
                        msg = email.message_from_bytes(response_part[1])
                        subject = msg.get("Subject", "")
                        if re.search(emailSubject, subject) or emailSubject.upper() in subject.upper():
                            matching_emails.append(e_id)
            if len(matching_emails)==0:
                if self.logger: self.logger.info("Aucun mail, n'a été trouvé.")
        else:
            matching_emails = email_ids
            if self.logger: self.logger.info(f"{len(email_ids)} mail(s) trouvé(s).")
        return matching_emails

    def read_email(self,email_id, get_attachment = False):
        if self.logger: self.logger.info(f"Tentative de lecture de l'email {email_id}.")
        imap_conn = self.__connection
        status, msg_data = imap_conn.fetch(email_id, "(RFC822)")
        if status != "OK":
            if self.logger: self.logger.error(f"Erreur lors de la lecture de l'email {email_id}.")
            raise
        if  msg_data[0]:
            raw_email = msg_data[0][1]
            msg = email.message_from_bytes(raw_email, policy=policy.default)
            if self.logger: self.logger.info(f"Contenu de l'email {email_id} correctement lu.")
            message=""
            for part in msg.walk():
                if part.get_content_type() in ["text/plain", "text/html"]:
                    raw = part.get_payload(decode=True)
                    encoding = chardet.detect(raw)["encoding"] or part.get_content_charset() or "utf-8"
                    if self.logger: self.logger.debug(f"Encodage détecté : {encoding} pour {part.get_content_type()}")
                    try:
                        message = raw.decode(encoding, errors='replace')
                    except LookupError:
                        message = raw.decode("iso-8859-1", errors='replace')
            if self.logger is not None and len(message) == 0: self.logger.info(f"Le contenu du mail est vide.")
            if not get_attachment:
                return {"message":message}
            else:
                data = self.__get_attachment(msg)
                return {"message": message, "attachement": data}
        else:
            if self.logger: self.logger.info(f"Le contenu du mail {email_id} est vide.")

    def __get_attachment(self, msg):
        lst = []
        if self.logger: self.logger.info(f"Récupération des pièces jointes.")
        for part in msg.walk():
            if part.get_content_disposition() == "attachment":
                filename = part.get_filename()
                if filename:
                    if self.logger: self.logger.info(f"Téléchargement de la pièce jointe : {filename}.")
                    raw = part.get_payload(decode=True)
                    mime_type, _ = mimetypes.guess_type(filename)
                    is_text = mime_type and mime_type.startswith("text")
                    binary = not is_text
                    encoding = None
                    content = raw
                    if is_text:
                        encoding = chardet.detect(raw)["encoding"]
                        if self.logger: self.logger.debug(f"Encodage détecté pour {filename}: {encoding}")
                        if encoding:
                            try:
                                content = raw.decode(encoding)
                                binary = False
                            except UnicodeDecodeError:
                                if self.logger: self.logger.error(
                                    f"Échec du décodage avec {encoding} pour {filename}. Enregistrement en binaire.")
                    lst.append({
                        "filename": filename,
                        "content": content,
                        "is_binary": binary,
                        "encoding": encoding if not binary else "binary"
                    })
        if self.logger and not lst: self.logger.info(f"Aucune pièce jointe n'a été trouvée.")
        return lst

    def logout(self):
        self.__connection.logout()


from abc import ABC, abstractmethod

class Connection(ABC):
    """
    Interface de base pour fournir un accès aux données.
    Permet de gérer la connexion à une ressource (base de données, fichier, etc.).
    """
    def __init__(self, resource_type: str, **kwargs):
        """
        Initialise la connexion.

        :param resource_type: Type de ressource (ex: 'sql', 'file').
        :param kwargs: Détails de connexion spécifiques à l'implémentation (serveur, utilisateur, mot de passe...).
        """
        self.resource_type = resource_type
        self.params = kwargs
        self.connection = None

    @abstractmethod
    def open(self):
        """Ouvre la connexion à la ressource."""
        pass

    @abstractmethod
    def close(self):
        """Ferme la connexion à la ressource."""
        pass

    @abstractmethod
    def commit(self):
        """Valide les changements (le cas échéant)."""
        pass

    @abstractmethod
    def is_open(self) -> bool:
        """Vérifie si la connexion est ouverte."""
        pass


import sqlalchemy as sa
from abc import ABC, abstractmethod

class DmlGenerator(ABC):
    """Abstract base class for DML Generator."""
    def __init__(self, engine: sa.engine.Engine):
        self.engine = engine

    def quote(self, name: str) -> str:
        """Quote database object names."""
        return self.engine.dialect.identifier_preparer.quote(name)

    @abstractmethod
    def to_json(self, columns) -> str:
        """Generate SQL for columns to json conversion."""
        pass

    @abstractmethod
    def to_sha2_256(self, string: str) -> str:
        """Generate SQL for SHA2-256 hash."""
        pass

    @abstractmethod
    def to_md5(self, string: str) -> str:
        """Generate SQL for MD5 hash."""
        pass

class MssqlDmlGenerator(DmlGenerator):
    """Concrete class for MSSQL DML Generator."""
    def to_json(self, columns) -> str:
        columns_expression = ", ".join([f"{str(col.expression)} AS {self.quote(col.name)}" for col in columns])
        sql_statement = f"(SELECT {columns_expression} FOR JSON PATH, WITHOUT_ARRAY_WRAPPER)"
        return sql_statement

    def to_sha2_256(self, string: str) -> str:
        sql_statement = f"CONVERT(varchar(64), HASHBYTES('SHA2_256', {string}), 2)"
        return sql_statement

    def to_md5(self, string: str) -> str:
        sql_statement = f"CONVERT(varchar(32), HASHBYTES('MD5', {string}), 2)"
        return sql_statement

class PostgresqlDmlGenerator(DmlGenerator):
    """Concrete class for PostgreSQL DML Generator."""
    def to_json(self, columns) -> str:
        columns_expression = ", ".join([f"{self.quote(col.name)}, {str(col.expression)}" for col in columns])
        sql_statement = f"json_build_object({columns_expression})::text"
        return sql_statement

    def to_sha2_256(self, string: str) -> str:
        sql_statement = f"encode(digest({string}, 'sha256'), 'hex')"
        return sql_statement

    def to_md5(self, string: str) -> str:
        sql_statement = f"md5({string})"
        return sql_statement

def createDmlGenerator(engine):
    """Factory function to create the appropriate DML Generator class."""
    dialect = engine.dialect.name
    if dialect == "mssql":
        return MssqlDmlGenerator(engine)
    elif dialect == "postgresql":
        return PostgresqlDmlGenerator(engine)
    else:
        raise NotImplementedError(f"DML Generator for database type '{dialect}' are not implemented.")


import os
import datetime
import polars as pl
from googleapiclient.discovery import build
from google.oauth2.credentials import Credentials
from google.auth.transport.requests import Request
from google_auth_oauthlib.flow import InstalledAppFlow

class GoogleCalendarConnector:
    def __init__(self, dictionnary: dict, logger):
        """
        Classe pour se connecter à l'API Google Calendar et récupérer des événements de calendrier.
        :param dictionnary: Un dictionnaire contenant les paramètres requis (comme les calendriers des collègues).
        :param logger: Un logger pour gérer les logs d'actions de la classe.
        """
        self.logger = logger
        self.scopes = dictionnary.get("scopes", ['https://www.googleapis.com/auth/calendar.readonly'])
        self.calendar_ids = dictionnary["calendar_ids"]
        self.token_file = dictionnary.get('token_file', 'token.json')
        self.credentials_file = dictionnary.get('credentials_file', 'credentials.json')
        self.service = None

    def __str__(self):
        """Retourne une représentation textuelle de l'objet GoogleCalendarConnector"""
        return f"Connexion Google Calendar avec les calendriers: {', '.join(self.calendar_ids)}"

    def connect(self):
        """Etablit la connexion à l'API Google Calendar."""
        try:
            self.logger.info("Tentative de connexion à Google Calendar...")
            creds = None
            if os.path.exists(self.token_file):
                creds = Credentials.from_authorized_user_file(self.token_file, self.scopes)
            if not creds or not creds.valid:
                if creds and creds.expired and creds.refresh_token:
                    creds.refresh(Request())
                else:
                    flow = InstalledAppFlow.from_client_secrets_file(self.credentials_file, self.scopes)
                    creds = flow.run_local_server(port=0)
                with open(self.token_file, 'w') as token:
                    token.write(creds.to_json())
            self.service = build('calendar', 'v3', credentials=creds)
            self.logger.info("Connexion à Google Calendar réussie.")
        except Exception as e:
            self.logger.error(f"Erreur lors de la connexion à Google Calendar: {e}")
            raise

    def get_events(self, max_results=10):
        """
        Récupère les événelents des calendriers des collègues spécifiés.
        :param max_results: Le nombre maximal d'événements à récupérer par calendrier
        :return: Un dictionnaire des événements pour chaque calendrier
        """
        try:
            now = datetime.datetime.now().isoformat() +  'Z'
            events_data = {}
            for calendar_id in self.calendar_ids:
                self.logger.info(f"Récupération des événements du calendrier: {calendar_id}")
                events_results = self.service.events().list(
                    calendarId=calendar_id,
                    timeMin=now,
                    maxResults=max_results,
                    singleEvents=True,
                    orderBy='startTime'
                ).execute()
                events = events_results.get('items', [])
                events_data[calendar_id] = events
                self.logger.info(f"{len(events)} événement(s) récupéré(s) pour {calendar_id}")
            return events_data
        except Exception as e:
            self.logger.error(f"Erreur lors de la récupération des événements: {e}")
            raise

    def save_events_to_csv(self, events_data, csv_filename='calendar_events.csv'):
        """
        Enregistre les événements dans un fichier csv en utilisant polars.
        :param events_data: Un dictionnaire contenant les événements par calendrier
        :param csv_filename: Le nom du csv
        """
        try:
            self.logger.info(f"Sauvegarde des événements dans le fichier CSV: {csv_filename}")
            data = {
                "Event Date": [],
                "Event Summary": [],
                "Owner Email": [],
                "Calendar Name": []
            }
            for calendar_id, events in events_data.items():
                for event in events:
                    data["Event Date"].append(event["start"].get("dateTime", event["start"].get("date")))
                    data["Event Summary"].append(event.get("summary", "No Summary"))
                    data["Owner Email"].append(event.get("organizer", {}).get('email', 'No Email Provided'))
                    data["Calendar Name"].append(event.get("organizer", {}).get("displayName", "No Name Provided"))
            df = pl.DataFrame(data)
            df.write_csv(csv_filename)
            self.logger.info(f"Les événements ont été sauvegardés dans {csv_filename}")
        except Exception as e:
            self.logger.error(f"Erreur lors de la sauvegarde des événements dans le CSV: {e}")
            raise


import psycopg2
import timeit
import polars as pl
from io import StringIO
from .timer import timer, get_timer
from .dataFactory import data_factory

class dbPgsql(data_factory):
    def __init__(self, dictionnary: dict, logger, batch_size=10_000):
        """
        instancie la classe dbMssql, qui hérite de dataFactory
        :param dictionnary: Un dictionnaire contenant tous les paramètres nécéssaires pour lancer une connexion postgre
        :param logger: un logger ads qui va gérer les logs des actions de la classe
        :param batch_size: la taille des batchs en lecture et écriture
        """
        self.connection = None
        self.logger = logger
        self.__database = dictionnary.get('database')
        self.__user = dictionnary.get('user')
        self.__password = dictionnary.get('password')
        self.__port = dictionnary.get('port')
        self.__host = dictionnary.get('host')
        self.batch_size = batch_size

    @timer
    def connect(self):
        """
        lance la connexion avec les identifiants passés à l'initialisation de la classe
        toutes les méthodes de la classe nécéssitent une connexion active
        :return: la connexion
        """
        if self.logger is not None: self.logger.info("Tentative de connexion avec la base de données.")
        try:
            self.connection = psycopg2.connect(
                database=self.__database,
                user=self.__user,
                password=self.__password,
                port=self.__port,
                host=self.__host
            )
            if self.logger is not None: self.logger.info(f"Connexion établie avec la base de données.")
            return self.connection
        except Exception as e:
            if self.logger is not None: self.logger.error(f"Échec de la connexion à la base de données: {e}")
            raise

    def sqlQuery(self, query: str, return_columns: bool = False):
        """
        lit la base de données avec la requête query
        :param return_columns: booléen qui indique si l'on veut récupérer les colonnes de la tables
        :param query: la requête
        :return: les données lues avec yield
        """
        self.logger.debug(f"Exécution de la requête de lecture: {query}")
        try:
            timer_start = timeit.default_timer()
            cpt_rows = 0
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                if return_columns:
                    cols = [desc for desc in cursor.description]
                    yield cols
                self.logger.info("Requête exécutée avec succès, début de la lecture des résultats.")
                while True:
                    rows = cursor.fetchmany(self.batch_size)
                    if not rows:
                        break
                    yield rows
                    cpt_rows+=len(rows)
                    self.logger.info(f"{cpt_rows} ligne(s) lue(s).")
            if get_timer():
                elapsed_time = timeit.default_timer() - timer_start
                self.logger.info(f"Temps d'exécution de sqlQuery: {elapsed_time:.4f} secondes")
        except Exception as e:
            self.logger.error(f"Échec de la lecture des données: {e}")
            raise

    @timer
    def sqlExec(self, query):
        """
        execute une requête sur la base de données, un create ou delete table par exemple
        :param query: la requête
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                self.connection.commit()
                self.logger.info(f"Requête exécutée avec succès.")
        except Exception as e:
            self.logger.error(f"Échec de l'exécution de la requête: {e}")
            raise

    @timer
    def sqlScalaire(self, query):
        """
        execute une requête et retourne le premier résultat
        :param query: la requête
        :return: le résultat de la requête
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query)
                result = cursor.fetchone()
                self.logger.info(f"Requête scalaire exécutée avec succès.")
                data = result[0] if result else None
                return data
        except Exception as e:
            self.logger.error(f"Échec de l'exécution de la requête: {e}")
            raise

    @timer
    def insert(self, schema: str, table: str, cols: [], row: []):
        """
        Insère des données dans la base de données
        :param schema: Nom du schema dans lequel est la table
        :param table: nom de la table dans laquelle insérer
        :param cols: liste des colonnes dans lesquelles insérer
        :param row: liste des valeurs à insérer
        :return: le résultat de l'opération, l'erreur et la ligne
        """
        placeholders = ', '.join(["%s"] * len(cols))
        table = f"{schema}.{table}" if schema else table
        query = f"INSERT INTO {table} ({', '.join(cols)}) VALUES ({placeholders})"
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeur(s) insérée(s) avec succès dans la table {table}.")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Échec de l'insertion des données: {e}")
            return "ERROR", str(e), row

    @timer
    def insertMany(self, schema: str, table: str, cols: [], rows: [[]]):
        """
        Insère des données par batch dans une table avec gestion des erreurs

        :param schema: Nom du schéma dans lequel se trouve la table
        :param table: Nom de la table dans laquelle insérer
        :param cols: Liste des colonnes dans lesquelles insérer
        :param rows: Liste des lignes à insérer
        :return: Le résultat de l'opération, l'erreur et la le batch concerné en cas d'erreur
        """
        failed_batches = []
        errors = []
        total_inserted = 0
        placeholders = ', '.join(["%s"] * len(cols))
        table = f"{schema}.{table}" if schema else table
        query = f"INSERT INTO {table} ({', '.join(cols)}) VALUES ({placeholders})"
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(f"Batch {batch_index}: {len(batch)} ligne(s) insérée(s) avec succès dans la table {table}. "
                                         f"Total inséré: {total_inserted}/{len(rows)} ligne(s).")
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
                return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows

    def __get_df(self, rows, cols):
        if isinstance(rows, list):
            df = pl.DataFrame(rows, schema=cols, orient='row', infer_schema_length=len(rows))
        elif isinstance(rows, pl.DataFrame):
            df = rows
        else:
            raise ValueError("Les données doivent être une liste de tuples ou un DataFrame polars")
        return df

    @timer
    def insertBulk(self, schema: str, table: str, cols: [], rows):
        """
        Insère des données dans une table avec la méthode rapide COPY (via copy_expert), en lots pour éviter les erreurs critiques.

        :param schema: Nom du schéma
        :param table: Nom de la table dans laquelle insérer
        :param cols: Liste des colonnes dans lesquelles insérer
        :param rows: Liste des lignes à insérer
        :return: Le résultat de l'opération, l'erreur et la le batch concerné en cas d'erreur
        """
        failed_batches = []
        errors = []
        total_inserted = 0
        try:
            df = self.__get_df(rows, cols)
            table = f"{schema}.{table}" if schema else table
            n_rows = df.shape[0]
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, n_rows, self.batch_size), start=1):
                    batch = df.slice(start, self.batch_size)
                    try:
                        buffer = StringIO()
                        batch.write_csv(buffer, include_header=False)
                        buffer.seek(0)
                        columns = ', '.join(cols)
                        query = f"COPY {table} ({columns}) FROM STDIN WITH CSV DELIMITER ','"
                        cursor.copy_expert(query, buffer)
                        self.connection.commit()
                        buffer.close()
                        total_inserted += batch.shape[0]
                        self.logger.info(f"Batch {batch_index}: {batch.shape[0]} ligne(s) insérée(s) avec succès dans la table {table}. "
                                         f"Total inséré: {total_inserted}/{n_rows} ligne(s).")
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'insertion du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
            return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'insertion des données: {e}")
            return "ERROR", str(e), rows

    @timer
    def upsert(self, schema: str, table: str, cols: [], row: [], conflict_cols: []):
        """
        Réalise une opération upsert sur la base
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param row: Liste des valeurs à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et la ligne en cas d'erreur
        """
        table = f"{schema}.{table}" if schema else table
        placeholders = ', '.join(["%s"] * len(cols))
        conflict_clause = ', '.join(conflict_cols)
        update_clause = ', '.join([f"{col} = EXCLUDED.{col}" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table} ({', '.join(cols)})
        VALUES ({placeholders})
        ON CONFLICT ({conflict_clause}) DO UPDATE
        SET {update_clause};
        """
        try:
            with self.connection.cursor() as cursor:
                cursor.execute(query, row)
                self.connection.commit()
                self.logger.info(f"{len(row)} valeurs insérée(s) ou mise(s) à jour dans la table {table}.")
                return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur lors de l'upsert dans la table {table}: {e}")
            return "ERROR", str(e), row

    @timer
    def upsertMany(self, schema: str, table: str, cols: [], rows: [[]], conflict_cols: []):
        """
        Réalise une opération upsert par batch sur la base
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param rows: Liste des lignes à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et le batch en cas d'erreur
        """
        failed_batches = []
        errors = []
        total_inserted = 0
        table = f"{schema}.{table}" if schema else table
        placeholders = ', '.join(["%s"] * len(cols))
        conflict_clause = ', '.join(conflict_cols)
        update_clause = ', '.join([f"{col} = EXCLUDED.{col}" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table} ({', '.join(cols)})
        VALUES ({placeholders})
        ON CONFLICT ({conflict_clause}) DO UPDATE
        SET {update_clause};
        """
        try:
            with self.connection.cursor() as cursor:
                for batch_index, start in enumerate(range(0, len(rows), self.batch_size), start=1):
                    batch = rows[start: start + self.batch_size]
                    try:
                        cursor.executemany(query, batch)
                        self.connection.commit()
                        total_inserted += len(batch)
                        self.logger.info(
                            f"Batch {batch_index}: {len(batch)} ligne(s) insérées ou mises à jour dans la table {table}. "
                            f"Total inséré ou mis à jour: {total_inserted}/{len(rows)} ligne(s)."
                        )
                    except Exception as batch_error:
                        self.connection.rollback()
                        self.logger.error(f"Erreur lors de l'upsert du batch {batch_index}: {batch_error}")
                        failed_batches.append(batch)
                        errors.append(str(batch_error))
                return ("ERROR" if failed_batches else "SUCCESS"), errors, failed_batches
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'upsert des données: {e}")
            return "ERROR", str(e), rows

    @timer
    def upsertBulk(self, schema: str, table: str, cols: [], rows, conflict_cols: []):
        """
        Réalise une opération upsert par batch sur la base avec copy_expert, en lots et avec l'aide d'une table temporaire
        :param schema: Le schéma de la table
        :param table: Le nom de la table
        :param cols: Liste des colonnes dans lequelles insérer
        :param rows: Liste des lignes à insérer
        :param conflict_cols: Colonnes à utiliser pour détecter les conflits
        :return: Le résultat de l'opération, l'erreur et le batch en cas d'erreur
        """
        temp_table = f"{table}_temp"
        table_full, temp_table_full = (f"{schema}.{table}", f"{schema}.{temp_table}") if schema else (table, temp_table)
        conflict_clause = ', '.join(conflict_cols)
        update_clause = ', '.join([f"{col} = EXCLUDED.{col}" for col in cols if col not in conflict_cols])
        query = f"""
        INSERT INTO {table_full} ({', '.join(cols)})
        SELECT {', '.join(cols)} FROM {temp_table_full}
        ON CONFLICT ({conflict_clause}) DO UPDATE
        SET {update_clause};
        """
        try:
            self.sqlExec(f"CREATE TEMP TABLE {temp_table_full} (LIKE {table_full} INCLUDING DEFAULTS)")
            result, errors, failed_batches = self.insertBulk(schema, temp_table, cols, rows)
            if result == "ERROR":
                return result, errors, failed_batches
            self.sqlExec(query)
            self.logger.info(f"{len(rows)} lignes insérée(s) ou mise(s) à jour dans la table {table}.")
            return "SUCCESS", [], []
        except Exception as e:
            self.connection.rollback()
            self.logger.error(f"Erreur critique lors de l'UPSERT : {e}")
            return "ERROR", str(e), rows
        finally:
            self.sqlExec(f"DROP TABLE IF EXISTS {temp_table_full}")
            self.logger.info(f"Table temporaire supprimée.")

    @timer
    def find_text_anywhere(self, search: str, include_views: bool = False, schema: str = None, table: str = None, log_every: int = 50):
        """
        Recherche une valeur textuelle dans toutes les colonnes textuelles de toutes les tables
        (et éventuellement vues) de tous les schémas.

        :param search: Texte à rechercher
        :param include_views: Inclure les vues dans la recherche (False par défaut)
        :param schema: Filtrer sur un schéma particulier
        :param table: Filtrer sur une table particulière
        :param log_every: Log de progression toutes les N colonnes
        """
        results = []
        base_sql = """
        SELECT c.table_schema, c.table_name, c.column_name
        FROM information_schema.columns c
        INNER JOIN information_schema.tables t
            ON c.table_schema = t.table_schema
            AND c.table_name = t.table_name
        WHERE c.data_type IN ('character varying', 'character', 'text')
        """
        params = []
        if not include_views:
            base_sql += " AND t.table_type = 'BASE TABLE'"
        if schema:
            base_sql += " AND c.table_schema = %s"
            params.append(schema)
        if table:
            base_sql += " AND c.table_name = %s"
            params.append(table)
        try:
            with self.connection.cursor() as cursor:
                self.logger.debug("Début de la recherche...")
                cursor.execute(base_sql, params)
                cols = cursor.fetchall()
                if not cols:
                    self.logger.warning("Aucune colonne trouvée correspondant aux critères.")
                    return results
                total = len(cols)
                total_tables = len(set((s, t) for s, t, _ in cols))
                total_schemas = len(set(s for s, _, _ in cols))
                self.logger.info(f"Analyse de {total} colonne(s) répartie(s) sur {total_tables} table(s) et {total_schemas} schéma(s)...")
                for i, (schema, table, column) in enumerate(cols, start=1):
                    full_table = f'{schema}.{table}'
                    query = f'SELECT COUNT(*) FROM {full_table} WHERE "{column}" ILIKE %s'
                    if i % log_every == 0 or i == total:
                        self.logger.debug(f"[{i}/{total}] Vérification de {full_table}.{column}")
                    try:
                        cursor.execute(query, (search,))
                        count = cursor.fetchone()[0]
                        if count > 0:
                            select_query = f'SELECT * FROM {full_table} WHERE "{column}" ILIKE \'{search}\''
                            results.append({
                                "schema": schema,
                                "table": table,
                                "column": column,
                                "count": count,
                                "query": select_query
                            })
                            self.logger.info(f"{count} ligne(s) trouvée(s) dans {full_table}.{column}")
                    except Exception as e:
                        self.logger.error(f"Erreur sur {full_table}.{column}: {e}")
            return results
        except Exception as e:
            self.logger.error(f'Erreur lors de la recherche du champ {search}: {e}')
            raise


import os
import dotenv

dotenv_file = dotenv.find_dotenv('../../.env')
dotenv.load_dotenv(dotenv_file)

pg_dwh_db = str(os.getenv('pg_dwh_db'))
pg_dwh_user = str(os.getenv('pg_dwh_user'))
pg_dwh_pwd = str(os.getenv('pg_dwh_pwd'))
pg_dwh_host = str(os.getenv('pg_dwh_host'))
pg_dwh_port = str(os.getenv('pg_dwh_port'))


mssql_dwh_db = str(os.getenv('mssql_dwh_db'))
mssql_dwh_user = str(os.getenv('mssql_dwh_user'))
mssql_dwh_pwd = str(os.getenv('mssql_dwh_pwd'))
mssql_dwh_host = str(os.getenv('mssql_dwh_host'))
mssql_dwh_port = str(os.getenv('mssql_dwh_port'))




from to_sort.Reforged.WriterHelper import WriterHelper

class SqlWriterHelper(WriterHelper):
    """
    Implémentation de WriterHelper pour une base SQL.
    """
    def guess_insert(self, target: str, columns: list):
        """
        Génère une requête d'insertion.
        :param target: Nom de la table cible.
        :param columns: Liste des colonnes.
        :return: Requête SQL d'insertion.
        """
        placeholders =  ", ".join(["%s"] * len(columns))
        col_names = ", ".join(columns)
        return f"INSERT INTO {target} ({col_names}) VALUES ({placeholders})"

    def guess_update(self, target: str, columns: list, conflict_cols: list):
        """
        Génère une requête de mise à jour.
        :param target: Nom de la table cible.
        :param columns: Liste des colonnes à mettre à jour.
        :param conflict_cols: Liste des colonnes pour détecter les conflits.
        :return: Requête SQL d'update.
        """
        update_clause = ", ".join([f"{col} = EXCLUDED.{col}" for col in columns if col not in conflict_cols])
        conflict_cols = ", ".join(conflict_cols)
        return f"""
        INSERT INTO {target} ({', '.join(columns)}
        VALUES ({', '.join(["%s"] * len(columns))}
        ON CONFLICT ({conflict_cols} DO UPDATE
        SET {update_clause}
        """

    def guess_if_exists_delete_and_create(self, target: str):
        """
        Génère les commandes pour supprimer et recréer une table SQL.
        :param target: Nom de la table.
        :return: Commandes SQL pour drop et create.
        """
        return [
            f"DROP TABLE IF EXISTS {target}",
            f"CREATE TABLE {target} (...)"
        ]


import sqlalchemy as sa
from sqlalchemy import exc
from abc import ABC, abstractmethod

class DdlOperations(ABC):
    """Abstract base class for DDL operations."""
    def __init__(self, engine: sa.engine.Engine, logger):
        self.engine = engine
        self.logger = logger

    def quote(self, name: str) -> str:
        """Quote database object names."""
        return self.engine.dialect.identifier_preparer.quote(name)

    def execute(self, query):
        """Execute a raw SQL query."""
        try:
            with self.engine.begin() as conn:
                conn.execute(sa.text(query))
                self.logger.debug(f"Executed query: {query}")
        except exc.SQLAlchemyError as e:
            self.logger.debug(f"Error executing query: {query} : {e}")
            raise

    def create_table(self, table: sa.Table) -> None:
        """Create a table."""
        if self.engine.dialect.has_table(self.engine.connect(), table.name, schema=table.schema):
            table.metadata.reflect(bind=self.engine, extend_existing=True, schema=table.schema, only=[table.name])
            self.logger.info(f"Table {table.schema}.{table.name} already exists.")
            return
        sql_statement = str(sa.schema.CreateTable(table).compile(self.engine))
        try:
            self.execute(sql_statement)
            table.metadata.reflect(bind=self.engine, extend_existing=True, schema=table.schema, only=[table.name])
            self.logger.info(f"Created table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error creating table {table.schema}.{table.name}: {e}.")
            raise

    def drop_table(self, table: sa.Table) -> None:
        """Drop a table."""
        if not self.engine.dialect.has_table(self.engine.connect(), table.name, schema=table.schema):
            self.logger.info(f"Table {table.schema}.{table.name} does not exist.")
            return
        sql_statement = str(sa.schema.DropTable(table).compile(self.engine))
        try:
            self.execute(sql_statement)
            self.logger.info(f"Dropped table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error dropping table {table.schema}.{table.name}: {e}.")
            raise

    def truncate_table(self, table: sa.Table) -> None:
        """Truncate a table."""
        if not self.engine.dialect.has_table(self.engine.connect(), table.name, schema=table.schema):
            self.logger.info(f"Table {table.schema}.{table.name} does not exist.")
            return
        sql_statement = f"TRUNCATE TABLE {self.quote(table.schema)}.{self.quote(table.name)}"
        try:
            self.execute(sql_statement)
            self.logger.info(f"Truncated table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error truncating table {table.schema}.{table.name}: {e}.")
            raise

    @abstractmethod
    def _add_column_sql(self, table: sa.Table, column: sa.Column) -> str:
        """Generate SQL for adding a column."""
        pass

    def add_column(self, table: sa.Table, column: sa.Column) -> None:
        """Add a column to a table."""
        if column.name in table.c:
            self.logger.info(f"Column {column.name} already exists in table {table.schema}.{table.name}.")
            return
        sql_statement = self._add_column_sql(table, column)
        try:
            self.execute(sql_statement)
            table.append_column(column)
            self.logger.info(f"Added column {column.name} to table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error adding column {column.name} to table {table.schema}.{table.name}: {e}.")
            raise

    @abstractmethod
    def _create_view_sql(self, view: sa.Table, definition: sa.Selectable, or_replace=True) -> str:
        """Generate SQL for creating a view."""
        pass

    def create_view(self, view: sa.Table, definition: sa.Selectable, or_replace=True) -> None:
        """Create a view."""
        if self.engine.dialect.has_table(self.engine.connect(), view.name, schema=view.schema) and not or_replace:
            view.metadata.reflect(bind=self.engine, views=True, extend_existing=True, schema=view.schema,
                                  only=[view.name])
            self.logger.info(f"View {view.schema}.{view.name} already exists")
            return
        sql_statement = self._create_view_sql(view, definition, or_replace)
        try:
            self.execute(sql_statement)
            view.metadata.reflect(bind=self.engine, views=True, extend_existing=True, schema=view.schema,
                                  only=[view.name])
            self.logger.info(f"Created view {view.schema}.{view.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(f"Error creating view {view.schema}.{view.name}: {e}.")
            raise

    def create_index(self, table, columns_name, index_name, unique=False):
        """Create an index on multiple table columns."""
        if index_name in [idx.name for idx in table.indexes]:
            self.logger.info(f"Index {index_name} already exists on table {table.schema}.{table.name}.")
            return
        for column_name in columns_name:
            if column_name not in table.c:
                raise ValueError(f"Column {column_name} does not exist in table {table.schema}.{table.name}.")
        index = sa.Index(index_name, *[table.c[column_name] for column_name in columns_name], unique=unique)
        try:
            index.create(self.engine)
            self.logger.info(
                f"Created index {index_name} on columns {', '.join(columns_name)} of table {table.schema}.{table.name}.")
        except exc.SQLAlchemyError as e:
            self.logger.error(
                f"Error creating index {index_name} on columns {', '.join(columns_name)} of table {table.schema}.{table.name}: {e}.")
            raise

class MssqlDdlOperations(DdlOperations):
    """Concrete class for MSSQL DDL operations."""
    def _add_column_sql(self, table: sa.Table, column: sa.Column) -> str:
        column_definition = f"{self.quote(column.name)} {column.type.compile(self.engine.dialect)}"
        sql_statement = f"ALTER TABLE {self.quote(table.schema)}.{self.quote(table.name)} ADD {column_definition}"
        return sql_statement

    def _create_view_sql(self, view: sa.Table, definition: sa.Selectable, or_replace=True) -> str:
        if or_replace:
            sql_statement = f"CREATE OR ALTER VIEW {self.quote(view.schema)}.{self.quote(view.name)} AS {definition.compile(compile_kwargs={'literal_binds': True})}"
        else:
            sql_statement = f"CREATE VIEW {self.quote(view.schema)}.{self.quote(view.name)} AS {definition.compile(compile_kwargs={'literal_binds': True})}"
        return sql_statement

class PostgresqlDdlOperations(DdlOperations):
    """Concrete class for PostgreSQL DDL operations."""
    def _add_column_sql(self, table: sa.Table, column: sa.Column) -> str:
        column_definition = f"{self.quote(column.name)} {column.type.compile(self.engine.dialect)}"
        sql_statement = f"ALTER TABLE {self.quote(table.schema)}.{self.quote(table.name)} ADD COLUMN {column_definition}"
        return sql_statement

    def _create_view_sql(self, view: sa.Table, definition: sa.Selectable, or_replace=True) -> str:
        if or_replace:
            sql_statement = f"CREATE OR REPLACE VIEW {self.quote(view.schema)}.{self.quote(view.name)} AS {definition.compile(compile_kwargs={'literal_binds': True})}"
        else:
            sql_statement = f"CREATE VIEW {self.quote(view.schema)}.{self.quote(view.name)} AS {definition.compile(compile_kwargs={'literal_binds': True})}"
        return sql_statement

def createDdlOperations(engine, logger):
    """Factory function to create the appropriate DDL operations class."""
    dialect = engine.dialect.name
    if dialect == "mssql":
        return MssqlDdlOperations(engine, logger)
    elif dialect == "postgresql":
        return PostgresqlDdlOperations(engine, logger)
    else:
        raise NotImplementedError(f"DDL operations for database type '{dialect}' are not implemented.")


import pymssql
import pytds
from to_sort.Reforged.Connection import Connection

class SqlServerConnection(Connection):
    """
    Implémentation de l'interface Connection pour SQL Server.
    """
    def __init__(self, driver: str = "pymssql", **kwargs):
        """
        Initialise la connexion SQL Server.
        :param driver: Pilote à utiliser ('pymssql' ou 'pytds').
        :param kwargs: Paramètres spécifiques à la connexion (serveur, utilisateur, etc.).
        """
        super().__init__(resource_type="sql", **kwargs)
        self.driver = driver

    def open(self):
        if self.connection:
            raise RuntimeError("La connexion est déjà ouverte.")
        try:
            if self.driver == "pymssql":
                self.connection = pymssql.connect(
                    server=self.params.get("server"),
                    user=self.params.get("user"),
                    password=self.params.get("password"),
                    database=self.params.get("database"),
                    port=self.params.get("port"),
                    charset=self.params.get("charset", "UTF-8")
                )
            elif self.driver == "pytds":
                self.connection = pytds.connect(
                    dsn=self.params.get("server"),
                    port=self.params.get("port"),
                    user=self.params.get("user"),
                    password=self.params.get("password"),
                    database=self.params.get("database"),
                    autocommit=self.params.get("autocommit", False)
                )
            else:
                raise ValueError(f"Pilote non supporté : {self.driver}")
        except Exception as e:
            raise RuntimeError(f"Erreur lors de l'ouverture de la connexion SQL Server : {e}")

    def close(self):
        if self.connection:
            self.connection.close()
            self.connection = None

    def commit(self):
        if self.connection:
            try:
                self.connection.commit()
            except Exception as e:
                raise RuntimeError(f"Erreur lors du commit SQL Server : {e}")
        else:
            raise RuntimeError("Impossible de commit : aucune connexion active.")

    def is_open(self) -> bool:
        return self.connection is not None and not self.connection.closed